from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, Union, Optional, List

from .enums import (
    ApplicationCommandOptionType,
    ApplicationCommandType,
    ChannelType,
)
from .channel import _channel_factory
from .permissions import Permissions
from .abc import Snowflake

if TYPE_CHECKING:
    from .types.interactions import (
        ApplicationCommand as ApplicationCommandPayload,
        ApplicationCommandInteractionDataOption as ApplicationCommandInteractionDataOptionPayload,
    )


ApplicationCommandOptionChoiceValue = Union[str, int]


__all__ = (
    "ApplicationCommandOptionChoice",
    "ApplicationCommandOption",
    "ApplicationCommand",
    "ApplicationCommandInteractionDataOption",
)


class ApplicationCommandInteractionDataOption:
    """
    A response from the API containing data from the user for a given interaction.
    This should not be created manually.

    .. versionadded:: 0.0.5

    Attributes
    -------------
    name: str
        The name of the parameter.
    type: ApplicationCommandOptionType
        The application command option type.
    value: Optional[str]
        The value passed into the command form the user.
    options: Optional[List[ApplicationCommandInteractionDataOption]]
        Present if this is a group or subcommand.
    focused: Optional[bool]
        True if this option is the currently focused option for autocomplete.
    """

    __slots__ = (
        "name",
        "type",
        "value",
        "options",
        "focused",
    )

    def __init__(self, data: ApplicationCommandInteractionDataOptionPayload):
        self.from_data(data)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}<name={self.name!r}, value={self.value!r}, type={self.type.name!r}>"

    def from_data(self, data: ApplicationCommandInteractionDataOptionPayload):
        self.name: str = data["name"]
        self.type: ApplicationCommandOptionType = ApplicationCommandOptionType(data["type"])
        self.value: Optional[str] = data.get("value")
        self.options: Optional[List[ApplicationCommandInteractionDataOption]] = [
            ApplicationCommandInteractionDataOption(i) for i in data.get("options", [])
        ] or None
        self.focused: Optional[bool] = data.get("focused")


class ApplicationCommandOptionChoice:
    """
    The possible choices that an application command can take.

    .. versionchanged:: 0.0.6

        The positional args are now kwargs.

    Parameters
    -----------
    name: :class:`str`
        The name of this option.
    value: :class:`str`
        The value given to this option.
    """

    def __init__(
            self,
            *,
            name: str,
            value: ApplicationCommandOptionChoiceValue = None):
        self.name: str = name
        self.value: Any = value if value is not None else name

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}<name={self.name!r}, value={self.value!r}>"

    @classmethod
    def from_data(cls, data: dict):
        return cls(
            name=data['name'],
            value=data['value'],
        )

    def to_json(self) -> dict:
        return {
            "name": self.name,
            "value": self.value,
        }


class ApplicationCommandOption:
    """
    An option displayed in a given application command.

    .. versionchanged:: 0.0.6

        The positional args are now kwargs.

    Parameters
    -----------
    name: :class:`str`
        The name of this option.
    type: :class:`ApplicationCommandOptionType`
        The type of this command option.
    description: :class:`str`
        The description given to this argument.
    required: Optional[:class:`bool`]
        Whether or not this option is required for the command to run.
    choices: List[:class:`ApplicationCommandOptionChoice`]
        A list of choices that this command can take.
    options: List[:class:`ApplicationCommandOption`]
        A list of options that go into the application command.
    channel_types: :class:`list`
        A list of channel types that the option accepts.

        .. versionadded:: 0.0.6
    autocomplete: :class:`bool`
        Whether or not the user typing this option should send autocomplete
        interaction payloads.

        .. versionadded:: 0.0.4
    name_localizations: Dict[:class:`str`, :class:`str`]
        A dictionary of language code to translation for the name of the command option.

        .. versionadded:: 0.0.6
    description_localizations: Dict[:class:`str`, :class:`str`]
        A dictionary of language code to translation for the name of the command option's description.

        .. versionadded:: 0.0.6
    min_value: Optional[:class:`int`]
        The minimum acceptable value of the field. Only usable for numeric types.

        .. versionadded:: 0.0.6
    max_value: Optional[:class:`int`]
        The maximum acceptable value of the field. Only usable for numeric types.

        .. versionadded:: 0.0.6

    Attributes
    -----------
    name: :class:`str`
        The name of this option.
    type: :class:`ApplicationCommandOptionType`
        The type of this command option.
    description: :class:`str`
        The description given to this argument.
    required: Optional[:class:`bool`]
        Whether or not this option is required for the command to run.
    choices: List[:class:`ApplicationCommandOptionChoice`]
        A list of choices that this command can take.
    options: List[:class:`ApplicationCommandOption`]
        A list of options that go into the application command.
    channel_types: :class:`list`
        A list of channel types that the option accepts.

        .. versionadded:: 0.0.4
    autocomplete: :class:`bool`
        Whether or not the user typing this option should send autocomplete
        interaction payloads.

        .. versionadded:: 0.0.4
    name_localizations: Dict[:class:`str`, :class:`str`]
        A dictionary of language code to translation for the name of the command option.

        .. versionadded:: 0.0.6
    description_localizations: Dict[:class:`str`, :class:`str`]
        A dictionary of language code to translation for the name of the command option's description.

        .. versionadded:: 0.0.6
    min_value: Optional[:class:`int`]
        The minimum acceptable value of the field. Only usable for numeric types.

        .. versionadded:: 0.0.6
    max_value: Optional[:class:`int`]
        The maximum acceptable value of the field. Only usable for numeric types.

        .. versionadded:: 0.0.6
    """

    def __init__(
            self,
            *,
            name: str,
            type: ApplicationCommandOptionType,
            description: str,
            required: bool = True,
            autocomplete: bool = False,
            name_localizations: Dict[str, str] = None,
            description_localizations: Dict[str, str] = None,
            choices: List[ApplicationCommandOptionChoice] = None,
            options: List[ApplicationCommandOption] = None,
            channel_types: List[ChannelType] = None,
            min_value: Optional[int] = None,
            max_value: Optional[int] = None,
            ):
        self.name: str = name
        self.type: ApplicationCommandOptionType = type
        self.description: str = description
        self.required: bool = required
        self.choices: List[ApplicationCommandOptionChoice] = choices or list()
        self.options: List[ApplicationCommandOption] = options or list()
        self.channel_types: list = channel_types or list()
        self.autocomplete: bool = autocomplete
        self.name_localizations = name_localizations or dict()
        self.description_localizations = description_localizations or dict()
        self.min_value = min_value
        self.max_value = max_value

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}<name={self.name!r}, type={self.type.name!r}, choices={self.choices!r}, options={self.options!r}>"

    def add_choice(self, choice: ApplicationCommandOptionChoice) -> None:
        """
        Add a choice to this instance.
        """

        self.choices.append(choice)

    def add_option(self, option: ApplicationCommandOption) -> None:
        """
        Add an option to this instance.
        """

        self.options.append(option)

    @classmethod
    def from_data(cls, data: dict):
        base_option = cls(
            name=data['name'],
            type=ApplicationCommandOptionType(data['type']),
            description=data['description'],
            name_localizations=data.get("name_localizations", dict()),
            description_localizations=data.get("description_localizations", dict()),
            required=data.get('required', False),
            autocomplete=data.get('autocomplete', False),
            min_value=data.get('min_value', None),
            max_value=data.get('max_value', None),
        )
        for choice in data.get('choices', list()):
            base_option.add_choice(ApplicationCommandOptionChoice.from_data(choice))
        for option in data.get('options', list()):
            base_option.add_option(cls.from_data(option))
        for channel_type in data.get('channel_types', list()):
            channel, _ = _channel_factory(channel_type)
            base_option.channel_types.append(channel)
        return base_option

    def to_json(self) -> dict:
        payload = {
            "name": self.name,
            "type": self.type.value,
            "description": self.description,
            "required": self.required,
            "autocomplete": self.autocomplete,
            "choices": [i.to_json() for i in self.choices],
            "options": [i.to_json() for i in self.options],
            "min_value": self.min_value,
            "max_value": self.max_value,
            "name_localizations": self.name_localizations,
            "description_localizations": self.description_localizations,
        }
        if self.type in [ApplicationCommandOptionType.subcommand, ApplicationCommandOptionType.subcommand_group]:
            payload.pop("required", None)
            payload.pop("choices", None)
            payload.pop("channel_types", None)
            payload.pop("autocomplete", None)
        else:
            payload["channel_types"] = [i.value for i in self.channel_types]
        return payload


class ApplicationCommand(Snowflake):
    """
    An instance of an application command.

    .. versionchanged:: 0.0.6

        The positional args are now kwargs.

    Attributes
    -----------
    name: :class:`str`
        The name of this command.
    description: Optional[:class:`str`]
        The description for this command.
    type: :class:`ApplicationCommandType`
        The type of the application command.
    options: List[:class:`ApplicationCommandOption`]
        A list of the options added to this command.
    name_localizations: Dict[:class:`str`, :class:`str`]
        A dictionary of language codes to translations for the name of the command.

        .. versionadded:: 0.0.6
    description_localizations: Dict[:class:`str`, :class:`str`]
        A dictionary of language codes to translations for the description of the command.

        .. versionadded:: 0.0.6
    id: :class:`int`
        The ID of this application command.
    application_id: :class:`int`
        The application ID that this command is attached to.
    dm_permissions: :class:`bool`
        Whether or not the command is runnable in DMs.

        .. versionadded:: 0.0.6
    default_member_permissions: Optional[:class:`discord.Permissions`]
        The default permissions needed to be able to run this command.

        .. versionadded:: 0.0.6

    Parameters
    -----------
    name: :class:`str`
        The name of this command.
    description: Optional[:class:`str`]
        The description for this command.
    type: :class:`ApplicationCommandType`
        The type of the application command. Defaults to a chat input (slash) command.
    options: List[:class:`ApplicationCommandOption`]
        A list of the options added to this command.
    name_localizations: Dict[:class:`str`, :class:`str`]
        A dictionary of language codes to translations for the name of the command.

        .. versionadded:: 0.0.6
    description_localizations: Dict[:class:`str`, :class:`str`]
        A dictionary of language codes to translations for the description of the command.

        .. versionadded:: 0.0.6
    dm_permissions: :class:`bool`
        Whether or not the command is runnable in DMs.

        .. versionadded:: 0.0.6
    default_member_permissions: :class:`discord.Permissions`
        The default permissions needed to be able to run this command.

        .. versionadded:: 0.0.6
    """

    def __init__(
            self,
            *,
            name: str,
            description: Optional[str] = None,
            type: ApplicationCommandType = ApplicationCommandType.chat_input,
            options: Optional[List[ApplicationCommandOption]] = None,
            name_localizations: Optional[Dict[str, str]] = None,
            description_localizations: Optional[Dict[str, str]] = None,
            dm_permissions: bool = True,
            default_member_permissions: Optional[Permissions] = None,
            ):
        self.name: str = name
        self.description: str = description or name
        self.type: ApplicationCommandType = type
        self.options: List[ApplicationCommandOption] = options or list()
        self.id: Optional[int] = None
        self.application_id: Optional[int] = None
        self.name_localizations = name_localizations or dict()
        self.description_localizations = description_localizations or dict()
        self.dm_permissions = dm_permissions
        self.default_member_permissions = default_member_permissions

    def __repr__(self):
        return f"{self.__class__.__name__}<name={self.name!r}, type={self.type.name!r}, options={self.options!r}, permissions={self.default_member_permissions!r}>"

    def add_option(self, option: ApplicationCommandOption):
        """
        Add an option to this command instance.
        """

        self.options.append(option)

    @classmethod
    def from_data(cls, data: ApplicationCommandPayload):
        command = cls(
            name=data['name'],
            description=data['description'],
            type=ApplicationCommandType(data.get('type', 1)),
            name_localizations=data.get("name_localizations", dict()),
            description_localizations=data.get("description_localizations", dict()),
            dm_permissions=data.get("dm_permissions", True),
            default_member_permissions=None if data.get("default_member_permissions") is None else Permissions(int(data["default_member_permissions"])),
        )
        command.id = int(data.get('id', 0)) or None
        command.application_id = int(data.get('application_id', 0)) or None
        for option in data.get('options', list()):
            command.add_option(ApplicationCommandOption.from_data(option))
        return command

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.to_json() == other.to_json()

    def to_json(self) -> ApplicationCommandPayload:
        v: ApplicationCommandPayload = {
            "name": self.name,
            "description": self.description,
            "type": self.type.value,
            "options": [i.to_json() for i in self.options],
            "name_localizations": self.name_localizations,
            "description_localizations": self.description_localizations,
            "default_member_permissions": str(self.default_member_permissions.value) if self.default_member_permissions else None,
            "dm_permissions": self.dm_permissions,
        }
        if self.type != ApplicationCommandType.chat_input:
            v.pop("options", None)
            v.pop("description", None)
            v.pop("description_localizations", None)
        return v



"""
The MIT License (MIT)

Copyright (c) 2021-present Kae Bartlett

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

import uuid
from typing import Optional, TYPE_CHECKING

from .models import InteractableComponent
from ..enums import TextStyle, ComponentType
from ..utils import MISSING

if TYPE_CHECKING:
    from ..types.components import (
        InputText as InputTextPayload,
    )


class InputText(InteractableComponent):
    """Represents an input text component.

    .. versionadded:: 0.0.5

    Attributes
    ------------
    label: :class:`str`
        The label you want to display above the input text component in the modal.
    style: :class:`discord.TextStyle`
        The style of the input text component.
    custom_id: Optional[:class:`str`]
        The ID that you want to assign to this input text component.
    placeholder: Optional[:class:`str`]
        The text shown (greyed out) in the input text component.
    min_length: Optional[:class:`int`]
        The minimum length of the required input.
    max_length: Optional[:class:`int`]
        The maximum length of the required input.
    required: Optional[:class:`bool`]
        Whether or not the text field is required.
    value: Optional[:class:`str`]
        The default value that goes into the text field.
    """

    __slots__ = ("label", "style", "custom_id", "min_length", "max_length",)
    TYPE = ComponentType.input_text

    def __init__(
            self,
            *,
            label: str = MISSING,
            custom_id: Optional[str] = None,
            style: TextStyle = TextStyle.short,
            placeholder: Optional[str] = None,
            min_length: Optional[int] = None,
            max_length: Optional[int] = None,
            required: bool = True,
            value: Optional[str] = None):
        self.label = label
        self.style = style
        self.placeholder = placeholder
        self.custom_id = custom_id or str(uuid.uuid1())
        self.min_length = min_length
        self.max_length = max_length
        self.required = required
        self.value = value

    def __repr__(self) -> str:
        attrs = (
            ('label', self.label),
            ('style', self.style),
            ('placeholder', self.placeholder),
            ('custom_id', self.custom_id),
            ('min_length', self.min_length),
            ('max_length', self.max_length),
            ('required', self.required),
            ('value', self.value),
        )
        inner = ' '.join('%s=%r' % t for t in attrs)
        return f'{self.__class__.__name__}({inner})'

    def to_dict(self) -> InputTextPayload:
        v = {
            "type": self.TYPE.value,
            "label": self.label,
            "style": self.style.value,
            "custom_id": self.custom_id,
            "placeholder": self.placeholder,
            "min_length": self.min_length,
            "max_length": self.max_length,
            "required": self.required,
            "value": self.value,
        }
        return v  # type: ignore

    @classmethod
    def from_dict(cls, data: InputTextPayload) -> InputText:
        """
        Construct an instance of an input text component from an API response.
        Because these aren't sent to bots, this is almost entirely useless.

        Parameters
        -----------
        data: :class:`dict`
            The payload data that the component should be constructed from.

        Returns
        -------
        :class:`discord.ui.InputText`
            The component that the payload describes.
        """

        return cls(
            style=TextStyle(data.get("style", TextStyle.short.value)),
            custom_id=data["custom_id"],
            label=data["label"],
            placeholder=data.get("placeholder"),
            min_length=data.get("min_length"),
            max_length=data.get("max_length"),
            required=data.get("required", False),
            value=data.get("value"),
        )



"""
The MIT License (MIT)

Copyright (c) 2021-present Kae Bartlett

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import Union, Tuple, List, Optional

from .models import LayoutComponent, ComponentHolder
from .button import Button, ButtonStyle
from ..enums import ComponentType
from ..components import _component_factory
from ..types.components import (
    MessageComponents as MessageComponentsPayload,
    ActionRow as ActionRowPayload,
)
from ..utils import MISSING


class ActionRow(LayoutComponent, ComponentHolder):
    """
    The main UI component for adding and ordering components on Discord
    messages.

    Parameters
    -----------
    components: Union[:class:`discord.ui.InteractableComponent`, :class:`discord.ui.LayoutComponent`]
        The components that you want to add to the row.
    """

    TYPE = ComponentType.action_row

    def to_dict(self) -> ActionRowPayload:
        return {
            "type": self.TYPE.value,
            "components": [i.to_dict() for i in self.components],
        }

    @classmethod
    def from_dict(cls, data: ActionRowPayload):
        components = data.get("components", list())
        new_components = []
        for i in components:
            v = _component_factory(i)
            new_components.append(v)
        return cls(*new_components)


class MessageComponents(ComponentHolder):
    """
    A set of components that are attached to a message.

    Parameters
    -----------
    components: :class:`discord.ui.LayoutComponent`
        The components that you want to add to the message.
    """

    def to_dict(self) -> MessageComponentsPayload:
        return [i.to_dict() for i in self.components]

    @classmethod
    def from_dict(cls, data: MessageComponentsPayload):
        new_components = []
        for i in data:
            v = _component_factory(i)
            new_components.append(v)
        return cls(*new_components)

    @classmethod
    def boolean_buttons(
            cls,
            yes_id: Optional[str] = None,
            no_id: Optional[str] = None,
            *,
            yes: Optional[Union[str, Tuple[str, Union[str, None]]]] = None,
            no: Optional[Union[str, Tuple[str, Union[str, None]]]] = None) -> MessageComponents:
        """
        Return a set of message components with yes/no buttons, ready for use. If provided, the given IDs
        will be used for the buttons. If not, the button custom IDs will be set to the strings
        "YES" and "NO".

        Parameters
        -----------
        yes: Optional[Union[str, Tuple[str, Union[str, None]]]]:
            Either a custom ID for the yes button, or two
            strings defining the button's text and custom ID respectively.
        no: Optional[Union[str, Tuple[str, Union[str, None]]]]:
            Either a custom ID for the no button, or two
            strings defining the button's text and custom ID respectively.

        Returns
        --------
        :class:`discord.ui.MessageComponents`
            A message components instance with the given buttons.
        """

        if yes:
            if isinstance(yes, str):
                yes = ("Yes", yes,)
        elif yes_id:
            yes = ("Yes", yes_id,)
        else:
            yes = ("Yes", None,)
        if no:
            if isinstance(no, str):
                no = ("No", no,)
        elif no_id:
            no = ("No", no_id,)
        else:
            no = ("No", None,)

        return cls(
            ActionRow(
                Button(label=yes[0], style=ButtonStyle.success, custom_id=yes[1]),
                Button(label=no[0], style=ButtonStyle.danger, custom_id=no[1]),
            ),
        )

    @classmethod
    def add_buttons_with_rows(cls, *buttons: Button):
        """
        Adds a list of buttons, breaking into a new :class:`ActionRow` automatically when it contains 5
        buttons. This does *not* check that you've added fewer than 5 rows.

        Parameters
        -----------
        buttons: :class:`discord.ui.Button`
            The buttons that you want to have added.
        """

        v = cls()
        while buttons:
            v.add_component(ActionRow(*buttons[:5]))
            buttons = buttons[5:]
        return v

    @classmethod
    def add_number_buttons(
            cls, numbers: List[int] = MISSING, *,
            add_negative: bool = False):
        """
        Creates a message components object with a list of number buttons added.

        Each number is added as its own button. Numbers provided as a list will be added as
        primary, where numbers added as negatives (if ``add_negative`` is set to ``True``)
        will be added as secondary buttons.

        A confirm button will not be automatically added.

        Parameters
        -----------
        numbers: List[:class:`int`]
            The numbers that you want to add. If not provided, then 1, 5, 10, 50, and 100 are used.
            A list of more than five values must not be given.
            These will be added as buttons with the custom ID ``NUMBER VALUE``, where ``VALUE``
            is the value shown on the button.
        add_negative: :class:`bool`
            Whether or not the negative versions of your numbers should be added as buttons.
            Defaults to ``False``.
        """

        if numbers is MISSING:
            numbers = [1, 5, 10, 50, 100]

        v = cls()
        if add_negative:
            v.add_component(ActionRow(*[
                Button(label=f"{i:+d}", custom_id=f"NUMBER {i}", style=ButtonStyle.primary)
                for i in numbers
            ]))
            v.add_component(ActionRow(*[
                Button(label=f"{-i:+d}", custom_id=f"NUMBER {-i}", style=ButtonStyle.secondary)
                for i in numbers
            ]))
        else:
            v.add_component(ActionRow(*[
                Button(label=str(i), custom_id=f"NUMBER {i}", style=ButtonStyle.primary)
                for i in numbers
            ]))
        return v



import argparse
import asyncio
import logging
import sys
import typing
import os
import importlib
import io
import traceback

import discord
from discord.ext import commands
import toml

from .cogs.utils.database import DatabaseWrapper
from .cogs.utils.redis import RedisConnection
from .cogs.utils.statsd import StatsdConnection
from .cogs.utils.custom_bot import Bot
from .cogs.utils.custom_context import PrintContext
from .cogs.utils.shard_manager import ShardManagerServer


class CascadingLogger(logging.getLoggerClass()):
    """
    A logger class that changes all of the handlers loglevels as well.
    Stdout will change to the loglevel set, and stderr will change to the max of what's
    been specified and WARNING.
    """

    def setLevel(self, level):
        for i in self.handlers:
            if isinstance(i, logging.StreamHandler):
                if i.stream.name == "<stdout>":
                    i.setLevel(level)
                elif i.stream.name == "<stderr>":
                    i.setLevel(max([level, logging.WARNING]))
        super().setLevel(level)


logging.setLoggerClass(CascadingLogger)
logger = logging.getLogger('vbu')


def set_log_level(
        logger_to_change: typing.Union[logging.Logger, str],
        log_level: str,
        minimum_level: int = None) -> None:
    """
    Set a logger to a default log level.

    Parameters
    -----------
    logger_to_change: :class`logging.Logger`
        The logger you want to change.
    log_level: :class:`str`
        The log level that you want to set the logger to.

    Raises
    -------
    :class:`ValueError`
        An invalid log_level was passed to the method.
    """

    # Make sure we're setting it to something
    if log_level is None:
        return

    # Get the logger we want to change
    if isinstance(logger_to_change, str):
        logger_to_change = logging.getLogger(logger_to_change)

    # Get the log level
    try:
        level = getattr(logging, log_level.upper())
    except AttributeError:
        raise ValueError(f"The log level {log_level.upper()} wasn't found in the logging module")

    # Set the level
    if minimum_level is not None:
        logger_to_change.setLevel(max([level, minimum_level]))
    else:
        logger_to_change.setLevel(level)


def validate_sharding_information(args: argparse.Namespace) -> typing.Optional[typing.List[int]]:
    """
    Validate the given shard information and make sure that what's passed in is accurate.

    Parameters
    -----------
    args: :class:`argparse.Namespace`
        The parsed argparse namespace for the program.

    Returns
    --------
    List[:class:`int`]
        A list of shard IDs to use with the bot.
    """

    # Set up some short vars for us to use
    set_min_and_max = args.min is not None and args.max is not None
    set_shardcount = args.shardcount is not None

    # If we haven't said anything, assume one shard
    if not set_shardcount and not set_min_and_max:
        args.shardcount = 1
        return [0]

    # If we haven't set a min or max but we HAVE set a shardcount,
    # then assume we're using all shards
    if set_shardcount and not set_min_and_max:
        args.min = 0
        args.max = args.shardcount - 1

    # If we gave a min and max but no shardcount, that's just invalid
    if set_min_and_max and not set_shardcount:
        logger.critical("You set a min/max shard amount but no shard count")
        exit(1)

    # Work out the shard IDs to launch with
    shard_ids = list(range(args.min, args.max + 1))
    return shard_ids


class LogFilter(logging.Filter):
    """
    Filters (lets through) all messages with level < LEVEL.

    To make our log levels work properly, we need to set up a new filter for our stream handlers
    We're going to send most things to stdout, but a fair few sent over to stderr
    """

    # Props to these folks who I stole all this from
    # https://stackoverflow.com/a/28743317/2224197
    # http://stackoverflow.com/a/24956305/408556

    def __init__(self, filter_level: int):
        self.filter_level = filter_level

    def filter(self, record):
        # "<" instead of "<=": since logger.setLevel is inclusive, this should
        # be exclusive
        return record.levelno < self.filter_level


def _set_default_log_level(logger_name, log_filter, formatter, loglevel):
    logger = logging.getLogger(logger_name) if isinstance(logger_name, str) else logger_name

    set_log_level(logger, 'DEBUG')

    stdout_logger = logging.StreamHandler(sys.stdout)
    stdout_logger.addFilter(log_filter)
    stdout_logger.setFormatter(formatter)
    set_log_level(stdout_logger, loglevel)  # type: ignore
    logger.addHandler(stdout_logger)

    stderr_logger = logging.StreamHandler(sys.stderr)
    stderr_logger.setFormatter(formatter)
    set_log_level(stderr_logger, loglevel, logging.WARNING)  # type: ignore
    logger.addHandler(stderr_logger)


def create_subclassed_loggers():
    DatabaseWrapper.logger = logging.getLogger("vbu.database")
    RedisConnection.logger = logging.getLogger("vbu.redis")
    StatsdConnection.logger = logging.getLogger("vbu.statsd")


def set_default_log_levels(args: argparse.Namespace) -> None:
    """
    Set the default levels for the logger.

    Parameters
    -----------
    bot: :class:`voxelbotutils.Bot`
        The custom bot object containing the logger, database logger, and redis logger.
    args: :class:`argparse.Namespace`
        The argparse namespace saying what levels to set each logger to.
    """

    # formatter = logging.Formatter('%(asctime)s [%(levelname)s][%(name)s] %(message)s')
    # formatter = logging.Formatter('{asctime} | {levelname: <8} | {module}:{funcName}:{lineno} - {message}', style='{')
    formatter = logging.Formatter('{asctime} | {levelname: <8} | {name}: {message}', style='{')
    log_filter = LogFilter(logging.WARNING)
    create_subclassed_loggers()
    loggers = [
        'vbu',
        'discord',
        'aiohttp',
        'aiohttp.access',
        'upgradechat',
    ]
    for i in loggers:
        if i is None:
            continue
        _set_default_log_level(i, log_filter, formatter, getattr(args, "loglevel", "ERROR"))


async def create_initial_database(db: DatabaseWrapper) -> bool:
    """
    Create the initial database using the internal database.psql file.
    """

    # Open the db file
    try:
        with open("./config/database.pgsql") as a:
            data = a.read()
    except Exception:
        return False

    # Get the statements
    create_table_statements = []
    current_line = ''
    for line in data.split('\n'):
        if line.lstrip().startswith('--'):
            continue
        current_line += line + '\n'
        if line.endswith(';') and not line.startswith(' '):
            create_table_statements.append(current_line.strip())
            current_line = ''

    # Let's do it baybeee
    for i in create_table_statements:
        if i and i.strip():
            await db(i.strip())

    # Sick we're done
    return True


async def start_database_pool(config: dict) -> None:
    """
    Start the database pool connection.
    """

    # Connect the database pool
    logger.info("Creating database pool")
    try:
        await DatabaseWrapper.create_pool(config['database'])
    except KeyError:
        raise Exception("KeyError creating database pool - is there a 'database' object in the config?")
    except ConnectionRefusedError:
        raise Exception(
            "ConnectionRefusedError creating database pool - did you set the right "
            "information in the config, and is the database running?"
        )
    except Exception:
        raise Exception("Error creating database pool")
    logger.info("Created database pool successfully")
    logger.info("Creating initial database tables")
    async with DatabaseWrapper() as db:
        await create_initial_database(db)


async def start_redis_pool(config: dict) -> None:
    """
    Start the redis pool connection.
    """

    # Connect the redis pool
    logger.info("Creating redis pool")
    try:
        await RedisConnection.create_pool(config['redis'])
    except KeyError:
        raise KeyError("KeyError creating redis pool - is there a 'redis' object in the config?")
    except ConnectionRefusedError:
        raise ConnectionRefusedError(
            "ConnectionRefusedError creating redis pool - did you set the right "
            "information in the config, and is the database running?"
        )
    except Exception:
        raise Exception("Error creating redis pool")
    logger.info("Created redis pool successfully")


class EventLoopCallbackHandler:
    """
    A callback handler for errors in the event loop so that they
    print to console properly, as well as sending to the event webhook
    if one is set up.

    Attributes
    -----------
    bot: :class:`voxelbotutils.Bot`
        The bot instance that the event loop is running.
    """

    bot: typing.Optional[Bot] = None

    @classmethod
    def callback(cls, future: asyncio.Future):
        # Print exceptions to console
        try:
            e = future.exception()
            if e is None:
                return
            raise e
        except (asyncio.CancelledError, asyncio.InvalidStateError):
            return
        except Exception as e:
            logger.error("Error in task was hit", exc_info=e)
            if cls.bot is None:
                return
            cls.bot.loop.create_task(cls.send_error_webhook(e))

    @classmethod
    async def send_error_webhook(cls, error: Exception):
        # Ping unhandled errors to the owners and to the event webhook
        error_string = "".join(traceback.format_exception(None, error, error.__traceback__))
        file_handle = io.StringIO(error_string + "\n")
        error_text = (
            f"Error `{error}` encountered.\nGuild `None`, channel `None`, "
            f"user `None`\n```\n[Task error]\n```"
        )

        # Assert some stuff
        try:
            assert cls.bot
            assert cls.bot.config
            assert cls.bot.user
        except AssertionError:
            return

        # DM to owners
        if cls.bot.config.get('dm_uncaught_errors', False):
            for owner_id in cls.bot.owner_ids:
                owner = cls.bot.get_user(owner_id) or await cls.bot.fetch_user(owner_id)
                file_handle.seek(0)
                file = discord.File(file_handle, filename="error_log.py")  # type: ignore
                await owner.send(error_text, file=file)

        # Ping to the webook
        event_webhook: typing.Optional[discord.Webhook] = cls.bot.get_event_webhook("unhandled_error")
        if not event_webhook:
            return
        try:
            avatar_url = str(cls.bot.user.display_avatar.url)
        except Exception:
            avatar_url = None
        try:
            username = cls.bot.user.name
        except Exception:
            username = cls.bot.application_id
            if username is None:
                username = cls.bot.config.get("oauth", {}).get("client_id", None) or "Application ID not found"
        if event_webhook:
            file_handle.seek(0)
            try:
                file = discord.File(file_handle, filename="error_log.py")
                await event_webhook.send(
                    error_text,
                    file=file,
                    username=f"{username} - Error",
                    avatar_url=avatar_url,
                    allowed_mentions=discord.AllowedMentions.none(),
                )
            except discord.HTTPException as e:
                cls.bot.logger.error(f"Failed to send webhook for event unhandled_error - {e}")


def set_event_loop():
    """
    Set up the event loop policy to use for asyncio, and set up
    a callback handler to log exceptions.
    """

    # Set up uvloop if we're on Linux
    try:
        import uvloop
        asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
    except ImportError:
        pass

    # If we're on Windows, set up a different event loop policy
    if sys.platform.startswith('win32'):
        if sys.version_info[0] == 3 and sys.version_info[1] >= 8:
            asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
        else:
            asyncio.set_event_loop(asyncio.ProactorEventLoop())

    # Set up the task factory
    def task_factory(loop: asyncio.AbstractEventLoop, coro) -> asyncio.Task:
        t = asyncio.Task(coro, loop=loop)
        t.add_done_callback(EventLoopCallbackHandler.callback)
        return t

    # And add it to our loop
    loop = asyncio.get_event_loop()
    loop.set_task_factory(task_factory)


def run_bot(args: argparse.Namespace) -> None:
    """
    Starts the bot, connects the database, runs the async loop forever.

    Parameters
    -----------
    args: :class:`argparse.Namespace`
        The arguments namespace that wants to be run.
    """

    os.chdir(args.bot_directory)
    set_event_loop()

    # And run file
    shard_ids = validate_sharding_information(args)
    bot = Bot(shard_count=args.shardcount, shard_ids=shard_ids, config_file=args.config_file)
    loop = bot.loop
    EventLoopCallbackHandler.bot = bot

    # Set up loggers
    bot.logger = logger.getChild("bot")
    set_default_log_levels(args)

    # Connect the database pool
    if bot.config.get('database', {}).get('enabled', False):
        db_connect_task = start_database_pool(bot.config)
        loop.run_until_complete(db_connect_task)

    # Connect the redis pool
    if bot.config.get('redis', {}).get('enabled', False):
        re_connect = start_redis_pool(bot.config)
        loop.run_until_complete(re_connect)

    # Load the bot's extensions
    logger.info('Loading extensions... ')
    bot.load_all_extensions()

    # Run the bot
    try:
        logger.info("Running bot")
        loop.run_until_complete(bot.start())
    except KeyboardInterrupt:
        logger.info("Logging out bot")
        loop.run_until_complete(bot.close())

    # We're now done running the bot, time to clean up and close
    if bot.config.get('database', {}).get('enabled', False):
        logger.info("Closing database pool")
        try:
            if DatabaseWrapper.pool:
                loop.run_until_complete(asyncio.wait_for(DatabaseWrapper.pool.close(), timeout=30.0))
        except asyncio.TimeoutError:
            logger.error("Couldn't gracefully close the database connection pool within 30 seconds")
    if bot.config.get('redis', {}).get('enabled', False):
        logger.info("Closing redis pool")
        RedisConnection.pool.close()

    logger.info("Closing asyncio loop")
    loop.stop()
    loop.close()


def run_interactions(args: argparse.Namespace) -> None:
    """
    Starts the bot, connects the database, runs the async loop forever.

    Parameters
    -----------
    args: :class:`argparse.Namespace`
        The arguments namespace that wants to be run.
    """

    from aiohttp.web import Application, AppRunner, TCPSite
    os.chdir(args.bot_directory)
    set_event_loop()

    # And run file
    bot = Bot(config_file=args.config_file, intents=discord.Intents.none())
    loop = bot.loop
    EventLoopCallbackHandler.bot = bot

    # Set up loggers
    bot.logger = logger.getChild("bot")
    set_default_log_levels(args)

    # Connect the database pool
    if bot.config.get('database', {}).get('enabled', False):
        db_connect_task = start_database_pool(bot.config)
        loop.run_until_complete(db_connect_task)

    # Connect the redis pool
    if bot.config.get('redis', {}).get('enabled', False):
        re_connect = start_redis_pool(bot.config)
        loop.run_until_complete(re_connect)

    # Load the bot's extensions
    logger.info('Loading extensions... ')
    bot.load_all_extensions()

    # Run the bot
    logger.info("Logging in bot")
    loop.run_until_complete(bot.login())
    websocket_task = None
    if args.connect:
        logger.info("Connecting bot to gateway")
        websocket_task = loop.create_task(bot.connect())

    # Create the webserver
    app = Application(loop=asyncio.get_event_loop(), debug=args.debug)
    app.router.add_routes(commands.get_interaction_route_table(bot, bot.config.get("pubkey", ""), path=args.path))

    # Start the HTTP server
    logger.info("Creating webserver...")
    application = AppRunner(app)
    loop.run_until_complete(application.setup())
    webserver = TCPSite(application, host=args.host, port=args.port)

    # Start the webserver
    loop.run_until_complete(webserver.start())
    logger.info(f"Server started - http://{args.host}:{args.port}/")

    # This is the forever loop
    try:
        logger.info("Running webserver")
        loop.run_forever()
    except KeyboardInterrupt:
        pass

    # We're now done running the webserver, time to clean up and close
    if websocket_task:
        websocket_task.cancel()
    if bot.config.get('database', {}).get('enabled', False):
        logger.info("Closing database pool")
        try:
            if DatabaseWrapper.pool:
                loop.run_until_complete(asyncio.wait_for(DatabaseWrapper.pool.close(), timeout=30.0))
        except asyncio.TimeoutError:
            logger.error("Couldn't gracefully close the database connection pool within 30 seconds")
    if bot.config.get('redis', {}).get('enabled', False):
        logger.info("Closing redis pool")
        RedisConnection.pool.close()

    logger.info("Closing asyncio loop")
    loop.stop()
    loop.close()


def run_website(args: argparse.Namespace) -> None:
    """
    Starts the website, connects the database, logs in the specified bots, runs the async loop forever.

    Parameters
    -----------
    args: :class:`argparse.Namespace`
        The arguments namespace that wants to be run.
    """

    # Load our imports here so we don't need to require them all the time
    from aiohttp.web import Application, AppRunner, TCPSite
    from aiohttp_jinja2 import setup as jinja_setup
    from aiohttp_session import setup as session_setup, SimpleCookieStorage
    from aiohttp_session.cookie_storage import EncryptedCookieStorage as ECS
    from jinja2 import FileSystemLoader
    import re
    import html
    from datetime import datetime as dt
    import markdown

    os.chdir(args.website_directory)
    set_event_loop()

    # Read config
    with open(args.config_file) as a:
        config = toml.load(a)

    # Create website object - don't start based on argv
    app = Application(loop=asyncio.get_event_loop(), debug=args.debug)
    app['static_root_url'] = '/static'
    for route in config['routes']:
        module = importlib.import_module(f"website.{route}", "temp")
        app.router.add_routes(module.routes)
    app.router.add_static('/static', os.getcwd() + '/website/static', append_version=True)

    # Add middlewares
    if args.debug:
        session_setup(app, SimpleCookieStorage(max_age=1_000_000))
    else:
        session_setup(app, ECS(os.urandom(32), max_age=1_000_000))
    jinja_env = jinja_setup(app, loader=FileSystemLoader(os.getcwd() + '/website/templates'))

    # Add our jinja env filters
    def regex_replace(string, find, replace):
        return re.sub(find, replace, string, re.IGNORECASE | re.MULTILINE)

    def escape_text(string):
        return html.escape(string)

    def timestamp(string):
        return dt.fromtimestamp(float(string))

    def int_to_hex(string):
        return format(hex(int(string))[2:], "0>6")

    def to_markdown(string):
        return markdown.markdown(string, extensions=['extra'])

    def display_mentions(string, users):
        def get_display_name(group):
            user = users.get(group.group('userid'))
            if not user:
                return 'unknown-user'
            return user.get('display_name') or user.get('username')
        return re.sub(
            '(?:<|(?:&lt;))@!?(?P<userid>\\d{16,23})(?:>|(?:&gt;))',
            lambda g: f'<span class="chatlog__mention">@{get_display_name(g)}</span>',
            string,
            re.IGNORECASE | re.MULTILINE,
        )

    def display_emojis(string):
        def get_html(group):
            return (
                f'<img class="discord_emoji" src="https://cdn.discordapp.com/emojis/{group.group("id")}'
                f'.{"gif" if group.group("animated") else "png"}" alt="Discord custom emoji: '
                f'{group.group("name")}" style="height: 1em; width: auto;">'
            )
        return re.sub(
            r"(?P<emoji>(?:<|&lt;)(?P<animated>a)?:(?P<name>\w+):(?P<id>\d+)(?:>|&gt;))",
            get_html,
            string,
            re.IGNORECASE | re.MULTILINE,
        )

    jinja_env.filters['regex_replace'] = regex_replace
    jinja_env.filters['escape_text'] = escape_text
    jinja_env.filters['timestamp'] = timestamp
    jinja_env.filters['int_to_hex'] = int_to_hex
    jinja_env.filters['markdown'] = to_markdown
    jinja_env.filters['display_mentions'] = display_mentions
    jinja_env.filters['display_emojis'] = display_emojis

    # Add our connections and their loggers
    app['database'] = DatabaseWrapper
    app['redis'] = RedisConnection
    app['logger'] = logger.getChild("route")
    app['stats'] = StatsdConnection

    # Add our config
    app['config'] = config

    loop = app.loop

    # Connect the database pool
    if app['config'].get('database', {}).get('enabled', False):
        db_connect_task = start_database_pool(app['config'])
        loop.run_until_complete(db_connect_task)

    # Connect the redis pool
    if app['config'].get('redis', {}).get('enabled', False):
        re_connect = start_redis_pool(app['config'])
        loop.run_until_complete(re_connect)

    # Add our bots
    app['bots'] = {}
    for index, (bot_name, bot_config_location) in enumerate(config.get('discord_bot_configs', dict()).items()):
        bot = Bot(f"./config/{bot_config_location}")
        app['bots'][bot_name] = bot
        if index == 0:
            set_default_log_levels(args)
        try:
            loop.run_until_complete(bot.login())
            bot.load_all_extensions()
        except Exception:
            logger.error(f"Failed to start bot {bot_name}", exc_info=True)
            exit(1)

    # Start the HTTP server
    logger.info("Creating webserver...")
    application = AppRunner(app)
    loop.run_until_complete(application.setup())
    webserver = TCPSite(application, host=args.host, port=args.port)

    # Start the webserver
    loop.run_until_complete(webserver.start())
    logger.info(f"Server started - http://{args.host}:{args.port}/")

    # This is the forever loop
    try:
        logger.info("Running webserver")
        loop.run_forever()
    except KeyboardInterrupt:
        pass

    # We're now done running the bot, time to clean up and close
    loop.run_until_complete(application.cleanup())
    if config.get('database', {}).get('enabled', False):
        logger.info("Closing database pool")
        try:
            if DatabaseWrapper.pool:
                loop.run_until_complete(asyncio.wait_for(DatabaseWrapper.pool.close(), timeout=30.0))
        except asyncio.TimeoutError:
            logger.error("Couldn't gracefully close the database connection pool within 30 seconds")
    if config.get('redis', {}).get('enabled', False):
        logger.info("Closing redis pool")
        RedisConnection.pool.close()

    logger.info("Closing asyncio loop")
    loop.stop()
    loop.close()


def run_sharder(args: argparse.Namespace) -> None:
    """
    Starts the sharder, connects the redis, runs the async loop forever.

    Parameters
    -----------
    args: :class:`argparse.Namespace`
        The arguments namespace that wants to be run.
    """

    set_event_loop()
    loop = asyncio.get_event_loop()
    set_default_log_levels(args)

    # Run the bot
    logger.info(f"Running sharder with {args.concurrency} shards")
    loop.create_task(ShardManagerServer(args.host, args.port, args.concurrency).run())
    try:
        loop.run_forever()
    except KeyboardInterrupt:
        logger.info("Logging out sharder")

    logger.info("Closing asyncio loop")
    loop.stop()
    loop.close()


def run_shell(args: argparse.Namespace) -> None:
    """
    Starts the shell for you.

    Parameters
    -----------
    args: :class:`argparse.Namespace`
        The arguments namespace that wants to be run.
    """

    os.chdir(args.bot_directory)
    set_event_loop()

    # And run file
    bot = Bot(config_file=args.config_file)
    loop = bot.loop

    # Set up loggers
    bot.logger = logger.getChild("bot")
    set_default_log_levels(args)

    # Connect the database pool
    if bot.config.get('database', {}).get('enabled', False):
        db_connect_task = start_database_pool(bot.config)
        loop.run_until_complete(db_connect_task)

    # Connect the redis pool
    if bot.config.get('redis', {}).get('enabled', False):
        re_connect = start_redis_pool(bot.config)
        loop.run_until_complete(re_connect)

    # Load the bot's extensions
    logger.info('Loading extensions... ')
    bot.load_all_extensions()

    # Set up the default env
    import voxelbotutils as vbu
    import discord
    from discord.ext import commands
    import textwrap
    import traceback
    import re
    env = {
        'bot': bot,
        'vbu': vbu,
        'discord': discord,
        'commands': commands,
        'ctx': PrintContext(bot),
    }

    # Run the bot
    try:
        logger.info("Running bot")
        loop.run_until_complete(bot.login())

        # Run our shell loop
        while True:

            # Get a user input
            line = input(">>> ")
            if not line.strip():
                line = "None"

            # See if they want to save that to a var
            var_name = None
            if (match := re.search(r"^([a-zA-Z_][a-zA-Z0-9_\.]*) ?=", line)):
                var_name = match.group(1)
                line = line.replace(match.group(0), "").lstrip()
            elif (match := re.search(r"^(?:from (?:[a-zA-Z_][a-zA-Z0-9_]*) )?import ([a-zA-Z_][a-zA-Z0-9_]*)", line)):
                var_name = match.group(1)
                line = line + f"\nreturn {var_name}"

            # Make sure that something is returned
            if not line.split("\n")[-1].startswith("return "):
                line = "return " + line

            # Make it asyncable
            code = f'async def _func():\n{textwrap.indent(line, "  ")}'

            # Run the function
            try:
                exec(code, env)
                func = env['_func']
                try:
                    ret = loop.run_until_complete(func())

                # Catch any errors
                except Exception:
                    print(traceback.format_exc().rstrip())

                # Deal with the result
                else:
                    if var_name is not None:
                        env.update({var_name: ret})
                    elif ret is not None:
                        print(repr(ret))
            except Exception:
                print(traceback.format_exc().rstrip())

    except KeyboardInterrupt:
        logger.info("Logging out bot")
        loop.run_until_complete(bot.close())

    # We're now done running the bot, time to clean up and close
    if bot.config.get('database', {}).get('enabled', False):
        logger.info("Closing database pool")
        try:
            if DatabaseWrapper.pool:
                loop.run_until_complete(asyncio.wait_for(DatabaseWrapper.pool.close(), timeout=30.0))
        except asyncio.TimeoutError:
            logger.error("Couldn't gracefully close the database connection pool within 30 seconds")
    if bot.config.get('redis', {}).get('enabled', False):
        logger.info("Closing redis pool")
        RedisConnection.pool.close()

    logger.info("Closing asyncio loop")
    loop.stop()
    loop.close()


def run_modify_commands(args: argparse.Namespace) -> None:
    """
    Modifies the commands available for the slash command instance

    Parameters
    -----------
    args: :class:`argparse.Namespace`
        The arguments namespace that wants to be run.
    """

    os.chdir(args.bot_directory)
    set_event_loop()

    # And run file
    bot = Bot(config_file=args.config_file)
    loop = bot.loop

    # Set up loggers
    bot.logger = logger.getChild("bot")
    set_default_log_levels(args)

    # Load the bot's extensions
    logger.info('Loading extensions... ')
    bot.load_all_extensions()

    # Run the bot
    logger.info("Running bot")
    loop.run_until_complete(bot.login())

    # Perform our action
    ctx = PrintContext(bot)
    if args.action == "add":
        command = bot.get_command("addapplicationcommands")
        assert command is not None
        ctx.command = command
        coro = ctx.invoke(ctx.command, args.guild)
    else:
        command = bot.get_command("removeapplicationcommands")
        assert command is not None
        ctx.command = command
        coro = ctx.invoke(ctx.command, args.guild)
    loop.run_until_complete(coro)

    # Logout the bot
    logger.info("Logging out bot")
    loop.run_until_complete(bot.close())

    logger.info("Closing asyncio loop")
    loop.stop()
    loop.close()



"""
The MIT License (MIT)

Copyright (c) 2021-present Kae Bartlett

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Optional, Union
from datetime import datetime as dt

from .mixins import Hashable
from .utils import MISSING, _get_as_snowflake, parse_time
from .enums import GuildScheduledEventPrivacyLevel, GuildScheduledEventStatus, GuildScheduledEventEntityType, try_enum
from .user import User

if TYPE_CHECKING:
    from .channel import VoiceChannel, StageChannel
    from .guild import Guild
    from .types.snowflake import Snowflake

__all__ = (
    "GuildScheduledEventEntityMetadata",
    "GuildScheduledEvent",
)


class GuildScheduledEventEntityMetadata:
    """
    The medatadata that's included for external guild scheduled events.

    .. versionadded:: 0.0.6

    Attributes
    ----------------
    location: :class:`str`
        The location that the guild event is scheduled to take place in.
    """

    def __init__(self, *, data):
        self.location: Optional[str] = data.get("location")

    def to_json(self) -> dict:
        return {
            "location": self.location,
        }

    def __repr__(self) -> str:
        attrs = (
            ('location', self.location),
        )
        inner = ' '.join('%s=%r' % t for t in attrs)
        return f'<{self.__class__.__name__} {inner}>'


class GuildScheduledEvent(Hashable):
    """
    An object representing a guild scheduled event.

    .. versionadded:: 0.0.6

    Attributes
    ------------
    id: :class:`int`
        The ID of the event.
    guild_id: :class:`int`
        The ID of the guild that the event is attached to.
    channel_id: Optional[:class:`int`]
        The ID of the channel where the event is going to occur.
    creator_id: :class:`int`
        The ID of the user who created the event. Will only be present for events
        created after October 25th 2021.
    name: :class:`str`
        The name of the event.
    description: :class:`str`
        The user-given description of the event.
    scheduled_start_time: :class:`datetime.datetime`
        The timestamp of when the event is meant to start.
    scheduled_end_time: Optional[:class:`datetime.datetime`]
        The timestamp of when the event is meant to end. Only required for external events.
    privacy_level: :class:`GuildScheduledEventPrivacyLevel`
        The privacy level of the event.
    status: :class:`GuildScheduledEventStatus`
        The event's status.
    entity_type: :class:`GuildScheduledEventEntityType`
        The entity that the event has been scheduled onto.
    entity_id: Optional[:class:`int`]
        The ID of the channel that the event has been scheduled onto.
    entity_metadata: Optional[:class:`GuildScheduledEventEntityMetadata`]
        The metadata attached to the event. Only applicable to external events.
    creator: Optional[:class:`User`]
        The creator of the event. Will only be present for events created after October 25th 2021.
    user_count: Optional[:class:`int`]
        The number of users who've said they're interested in the event.
    """

    __slots__ = (
        "_state",
        "id",
        "guild_id",
        "channel_id",
        "creator_id",
        "name",
        "description",
        "scheduled_start_time",
        "scheduled_end_time",
        "privacy_level",
        "status",
        "entity_type",
        "entity_id",
        "entity_metadata",
        "creator",
        "user_count",
    )

    def __init__(self, *, state, data):
        self._state = state
        self.id: Snowflake = _get_as_snowflake(data, "id")
        self.guild_id: Optional[Snowflake] = _get_as_snowflake(data, "guild_id")
        self.channel_id: Optional[Snowflake] = _get_as_snowflake(data, "channel_id")
        self.creator_id: Optional[Snowflake] = _get_as_snowflake(data, "creator_id")
        self.name: str = data.get("name")
        self.description: Optional[str] = data.get("description")
        self.scheduled_start_time: dt = parse_time(data.get("scheduled_start_time"))
        self.scheduled_end_time: Optional[dt] = parse_time(data.get("scheduled_end_time"))
        self.privacy_level: GuildScheduledEventPrivacyLevel = try_enum(GuildScheduledEventPrivacyLevel, data.get("privacy_level"))
        self.status: GuildScheduledEventStatus = try_enum(GuildScheduledEventStatus, data.get("status"))
        self.entity_type: GuildScheduledEventEntityType = try_enum(GuildScheduledEventEntityType, data.get("entity_type"))
        self.entity_id: Optional[Snowflake] = _get_as_snowflake(data, "entity_id")
        self.entity_metadata: Optional[GuildScheduledEventEntityMetadata] = None
        if data.get("entity_metadata"):
            self.entity_metadata = GuildScheduledEventEntityMetadata(data=data.get("entity_metadata"))
        self.creator: Optional[User] = None
        if data.get("creator"):
            self.creator = User(state=state, data=data.get("creator"))
        self.user_count: Optional[int] = data.get("user_count")

    def __repr__(self) -> str:
        attrs = (
            ('id', self.id),
            ('name', self.name),
        )
        inner = ' '.join('%s=%r' % t for t in attrs)
        return f'<{self.__class__.__name__} {inner}>'

    @property
    def guild(self) -> Optional[Guild]:
        """:class:`Guild`: The guild this event belongs to."""

        return self._state._get_guild(self.guild_id)

    @property
    def channel(self) -> Optional[Union[VoiceChannel, StageChannel]]:
        """Optional[Union[:class:`VoiceChannel`, :class:`StageChannel`]]: The channel where the event will occur."""

        if self.channel_id:
            return self._state._get_channel(self.channel_id)
        return None

    async def edit(self, **kwargs):
        """
        Edit the attributes of this event.

        Parameters
        -------------
        channel: Optional[Union[:class:`VoiceChannel`, :class:`StageChannel`]]
            The channel where the event is going to occur. Can only be set to ``None`` if
            the event is occuring externally.
        entity_metadata: Optional[:class:`GuildScheduledEventEntityMetadata`]
            The metadata of the event's occurence.
        name: :class:`str`
            The name that you want to set the event to.
        privacy_level: :class:`GuildScheduledEventPrivacyLevel`
            The privacy level that you want to change the event to.
        scheduled_start_time: :class:`datetime.datetime`
            The time that the event is meant to start.
        scheduled_end_time: :class:`datetime.datetime`
            The time that the event is meant to end.
        description: :class:`str`
            The description shown for the event.
        entity_type: :class:`GuildScheduledEventEntityType`
            The entity type for the event.
        status: :class:`GuildScheduledEventStatus`
            The status for the scheduled event. Once an event is completed or
            cancelled, you can no longer update its status.

        Raises
        ------
        Forbidden
            You do not have permissions to edit the event.
        NotFound
            The event was deleted already.
        HTTPException
            Editing the event failed.

        Returns
        --------
        :class:`.GuildScheduledEvent`
            The newly edited event.
        """

        payload = {}

        name = kwargs.pop("name", MISSING)
        if name is not MISSING:
            payload["name"] = name

        entity_metadata = kwargs.pop("entity_metadata", MISSING)
        if entity_metadata is not MISSING:
            payload["entity_metadata"] = entity_metadata.to_json()

        channel = kwargs.pop("channel", MISSING)
        if channel is not MISSING:
            payload["channel_id"] = channel.id

        privacy_level = kwargs.pop("privacy_level", MISSING)
        if privacy_level is not MISSING:
            payload["privacy_level"] = privacy_level.value

        scheduled_start_time = kwargs.pop("scheduled_start_time", MISSING)
        if scheduled_start_time is not MISSING:
            payload["scheduled_start_time"] = scheduled_start_time.isoformat()

        scheduled_end_time = kwargs.pop("scheduled_end_time", MISSING)
        if scheduled_end_time is not MISSING:
            if scheduled_end_time:
                payload["scheduled_end_time"] = scheduled_end_time.isoformat()
            else:
                payload["scheduled_end_time"] = scheduled_end_time

        description = kwargs.pop("description", MISSING)
        if description is not MISSING:
            payload["description"] = description

        entity_type = kwargs.pop("entity_type", MISSING)
        if entity_type is not MISSING:
            payload["entity_type"] = entity_type.value

        status = kwargs.pop("status", MISSING)
        if status is not MISSING:
            payload["status"] = status.value

        returned = await self._state.http.modify_guild_scheduled_event(self.guild_id, self.id, payload)
        if returned:
            return self.__class__(state=self._state, data=returned)
        return None

    async def delete(self) -> None:
        """
        Delete this event from the guild.

        Raises
        ------
        Forbidden
            You do not have proper permissions to delete the event.
        NotFound
            The event was deleted already.
        HTTPException
            Deleting the event failed.
        """

        await self._state.http.delete_guild_scheduled_event(self.guild_id, self.id)

    async def cancel(self) -> None:
        """
        Set the status of the event to canceled.

        Raises
        ------
        Forbidden
            You do not have proper permissions to edit the event.
        NotFound
            The event was deleted already.
        HTTPException
            Editing the event failed.
        """

        payload = {
            "status": GuildScheduledEventStatus.canceled.value,
        }
        await self._state.http.modify_guild_scheduled_event(self.guild_id, self.id, payload)

    def __repr__(self) -> str:
        attrs = (
            ('id', self.id),
            ('guild_id', self.guild_id),
            ('name', self.name),
        )
        inner = ' '.join('%s=%r' % t for t in attrs)
        return f'<{self.__class__.__name__} {inner}>'



from __future__ import annotations

from .button import ButtonStyle, Button
from .select_menu import SelectMenu, SelectOption, UserSelectMenu, RoleSelectMenu, MentionableSelectMenu, ChannelSelectMenu
from .models import BaseComponent, DisableableComponent, ComponentHolder, InteractedComponent, InteractableComponent, LayoutComponent
from .action_row import ActionRow, MessageComponents
from .modal import Modal
from .input_text import InputText


__all__ = (
    "ButtonStyle",
    "SelectMenu",
    "UserSelectMenu",
    "RoleSelectMenu",
    "MentionableSelectMenu",
    "ChannelSelectMenu",
    "BaseComponent",
    "InteractableComponent",
    "LayoutComponent",
    "ActionRow",
    "Button",
    "SelectOption",
    "DisableableComponent",
    "MessageComponents",
    "Modal",
    "ComponentHolder",
    "InputText",
    "InteractedComponent",
)



# -*- coding: utf-8 -*-

"""
The MIT License (MIT)

Copyright (c) 2015-2021 Rapptz
Copyright (c) 2021-present Kae Bartlett

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import Any, Dict, Generic, List, Optional, TYPE_CHECKING, Tuple, Union, TypeVar
import asyncio
import json

import aiohttp
from aiohttp import web

from . import utils
from .application_commands import ApplicationCommandInteractionDataOption
from .enums import try_enum, InteractionType, InteractionResponseType
from .errors import InteractionResponded, HTTPException, ClientException, InvalidData
from .channel import PartialMessageable, ChannelType, _threaded_channel_factory
from .user import User
from .member import Member
from .role import Role
from .message import Message, Attachment
from .object import Object
from .permissions import Permissions
from .webhook.async_ import async_context, Webhook, handle_message_parameters
from .ui.models import InteractedComponent

__all__ = (
    'Interaction',
    'InteractionMessage',
    'InteractionResponse',
    'InteractionResolved',
)

if TYPE_CHECKING:
    from aiohttp import ClientSession

    from .types.interactions import (
        Interaction as InteractionPayload,
        InteractionData,
    )
    from .types.snowflake import Snowflake
    from .application_commands import ApplicationCommandOptionChoice
    from .guild import Guild
    from .state import ConnectionState
    from .file import File
    from .mentions import AllowedMentions
    from .embeds import Embed
    from .channel import VoiceChannel, StageChannel, TextChannel, CategoryChannel, StoreChannel, PartialMessageable
    from .threads import Thread
    from .ui.action_row import MessageComponents
    from .ui.modal import Modal

    InteractionChannel = Union[
        VoiceChannel, StageChannel, TextChannel, CategoryChannel, StoreChannel, Thread, PartialMessageable,
    ]

MISSING: Any = utils.MISSING


T = TypeVar("T", None, str)


class InteractionResolved:
    """
    A set of resolved data from an interaction.

    .. versionadded:: 0.0.3

    .. versionchanged:: 0.0.7

        Changed all list attributes to dicts.

    Attributes
    ------------
    users: Dict[:class:`int`, Union[:class:`User`, :class:`Member`]]
        The users who were mentioned in the interaction.
    members: Dict[:class:`int`, :class:`Member`]
        The members who were mentioned in the interaction.
    roles: Dict[:class:`int`, :class:`Role`]
        The roles that were mentioned in the interaction.
    channels: Dict[:class:`int`, Union[:class:`VoiceChannel`, :class:`StageChannel`, :class:`TextChannel`, :class:`CategoryChannel`, :class:`StoreChannel`, :class:`Thread`]]
        The channels that were mentioned in the interaction.
    messages: Dict[:class:`int`, :class:`Message`]
        The messages that were mentioned in the interaction.
    attachments: Dict[:class:`int`, :class:`Attachment`]
        The attachments that were sent with the payload.

        .. versionadded:: 0.0.7
    """

    __slots__ = (
        "_interaction",
        "_state",
        "_cs_users",
        "_cs_members",
        "_cs_roles",
        "_cs_channels",
        "_cs_messages",
        "_cs_attachments",
        "_data",
    )

    def __init__(self, *, interaction, data: dict, state: ConnectionState):
        self._interaction: Interaction = interaction
        self._state = state
        self._data = data

    @utils.cached_slot_property("_cs_users")
    def users(self) -> Dict[int, Union[User, Member]]:
        users = self.members.copy()
        user_data = self._data.get("users", dict())
        member_data = self._data.get("members", dict())
        for uid, d in member_data.items():
            if int(uid) not in users:
                d.update({"user": user_data.pop(uid)})
        users.update({
            int(i): User(data=d, state=self._state)
            for i, d in user_data.items()
        })
        return users

    @utils.cached_slot_property("_cs_members")
    def members(self) -> Dict[int, Member]:
        members = {}
        if not self._interaction.guild:
            return {}
        user_data = self._data.get("users", dict())
        member_data = self._data.get("members", dict())
        for uid, d in member_data.items():
            d.update({"user": user_data.pop(uid)})
        members.update({
            int(i): Member(data=d, state=self._state, guild=self._interaction.guild)
            for i, d in member_data.items()
        })
        return members

    @utils.cached_slot_property("_cs_roles")
    def roles(self) -> Dict[int, Role]:
        roles = {}
        if self._interaction.guild:
            for i, d in self._data.get("roles", dict()).items():
                roles[int(i)] = Role(guild=self._interaction.guild, state=self._state, data=d)
        return roles

    @utils.cached_slot_property("_cs_channels")
    def channels(self) -> Dict[int, InteractionChannel]:
        channels = {}
        for _, d in self._data.get("channels", dict()).items():
            factory, ch_type = _threaded_channel_factory(d['type'])
            if factory is None:
                raise InvalidData('Unknown channel type {type} for channel ID {id}.'.format_map(d))
            if ch_type in (ChannelType.group, ChannelType.private):
                channel = factory(me=self.user, data=d, state=self._connection) # type: ignore
            else:
                guild = self._interaction.guild
                channel = factory(guild=guild, state=self._connection, data=d) # type: ignore
            channels[channel.id] = channel
        return channels

    @utils.cached_slot_property("_cs_messages")
    def messages(self) -> Dict[int, Message]:
        messages = {}
        for i, d in self._data.get("messages", dict()).items():
            messages[int(i)] = Message(state=self._state, channel=self._interaction.channel, data=d)
        return messages

    @utils.cached_slot_property("_cs_attachments")
    def attachments(self) -> Dict[int, Attachment]:
        attachments = {}
        for i, d in self._data.get("attachments", dict()).items():
            attachments[int(i)] = Attachment(state=self._state, data=d)
        return attachments


class Interaction(Generic[T]):
    """Represents a Discord interaction.

    An interaction happens when a user does an action that needs to
    be notified. Current examples are slash commands and components.

    Attributes
    -----------
    id: :class:`int`
        The interaction's ID.
    type: :class:`InteractionType`
        The interaction type.
    guild_id: Optional[:class:`int`]
        The guild ID the interaction was sent from.
    channel_id: Optional[:class:`int`]
        The channel ID the interaction was sent from.
    application_id: :class:`int`
        The application ID that the interaction was for. Unless you're running multiple
        bots in one script, this will be generally useless for most people.
    user: Optional[Union[:class:`User`, :class:`Member`]]
        The user or member that sent the interaction.
    message: Optional[:class:`Message`]
        The message that this interaction was spawned from. Only present for component
        interactions.
    component: Optional[:class:`ui.InteractableComponent`]
        The component that was interacted with to spawn this interaction.
        This will be ``None`` if the interaction was created from an application command.
    values: Optional[List[:class:`str`]]
        The values that were passed back from the interaction. If this interaction
        did not give back any values then this will be ``None``. This is different from
        the values being an empty list - that would be the user did not
        provide any values for a valid component.

        .. versionchanged:: 0.0.5
            This is now a list of :class:`str` objecs instead of ``dict``.
    token: :class:`str`
        The token to continue the interaction. These are valid
        for 15 minutes.
    data: :class:`dict`
        The data from the interaction. You don't need to use this unless you're building your own
        interaction handler, or dealing with an interaction that isn't yet added to the library.
    resolved: :class:`InteractionResolved`
        The resolved interaction data.
    options: Optional[List[:class:`ApplicationCommandInteractionDataOption`]]
        A list of options passed in from the interaction. For commands, this will be the options of
        the command; for autocompletes this will be a list of the options (as filled by the user) for the
        command.

        .. versionadded:: 0.0.4

        .. versionchanged:: 0.0.5
            Now is a list of :class:`ApplicationCommandInteractionDataOption` objects instead of a
            raw dictionary.
    components: Optional[List[:class:`ui.InteractionComponent`]]
        The components that are returned with the interaction. This is only used with modals so far.
        This skips the base level component (ie the modal object) and only gives its components.

        .. versionadded:: 0.0.5
    custom_id: Optional[:class:`str`]
        The custom ID associated with the main component of this interaction.

        .. versionadded:: 0.0.5
    user_locale: :class:`str`
        The locale of the user's client.

        .. versionadded:: 0.0.6
    guild_locale: Optional[:class:`str`]
        The locale of the guild where the interaction was invoked. Will be ``None`` if the interaction
        was invoked from a DM.

        .. versionadded:: 0.0.6
    locale: :class:`str`
        Returns the user locale or the guild locale

        .. versionadded:: 0.0.6
    """

    __slots__: Tuple[str, ...] = (
        'id',
        'type',
        'guild_id',
        'channel_id',
        'data',
        'component',
        'components',
        'values',
        'application_id',
        'target_id',
        'message',
        'user',
        'token',
        'version',
        'custom_id',
        '_cs_resolved',
        '_permissions',
        '_state',
        '_session',
        '_original_message',
        '_cs_response',
        '_cs_followup',
        '_cs_channel',
        '_cs_command_name',
        'options',
        'user_locale',
        'guild_locale',
    )

    def __init__(self, *, data: InteractionPayload, state: ConnectionState):
        self._state: ConnectionState = state
        self._session: ClientSession = state.http._HTTPClient__session  # type: ignore
        self._original_message: Optional[InteractionMessage] = None
        self._from_data(data)

    def _from_data(self, payload: InteractionPayload):
        self.id: Snowflake = int(payload['id'])
        self.type: InteractionType = try_enum(InteractionType, payload['type'])
        self.token: str = payload['token']
        self.version: int = payload['version']
        self.channel_id: Optional[int] = utils._get_as_snowflake(payload, 'channel_id')
        self.guild_id: Optional[int] = utils._get_as_snowflake(payload, 'guild_id')
        self.application_id: int = int(payload['application_id'])
        self.target_id = payload.get("target_id")

        # Parse the message object
        self.message: Optional[Message]
        try:
            self.message = Message(state=self._state, channel=self.channel, data=payload['message'])  # type: ignore
        except KeyError:
            self.message = None

        # The data that the interaction gave back to us - not optional, but documented as optional
        # This contains all the data ABOUT the interaction that's given back to us
        self.data: Optional[InteractionData] = payload.get('data')

        # Parse the component that triggered the interaction - this does NOT apply to modals
        self.component: Optional[InteractedComponent] = None
        try:
            if self.message:
                self.component = self.message.components.get_component(self.data['custom_id'])  # type: ignore
        except (KeyError, AttributeError):
            pass

        # Parse the main custom ID
        self.custom_id: T = None
        if self.data:
            self.custom_id = self.data.get('custom_id')

        # Parse the given values from the component - this is only used by select components
        self.values: Optional[List[str]] = None
        if self.data and 'values' in self.data:
            self.values = self.data['values']

        # Parse the returned options from the user - this is used by all application commands (including autocorrect)
        self.options: Optional[List[ApplicationCommandInteractionDataOption]] = None
        if self.data and 'options' in self.data:
            self.options = [ApplicationCommandInteractionDataOption(i) for i in self.data['options']]

        # Parse the returned components - this is used by modals
        self.components: Optional[List[InteractedComponent]] = None
        if self.data and 'components' in self.data:
            self.components = [InteractedComponent.from_data(i) for i in self.data['components']]

        # Parse the user and their permissions
        self.user: Optional[Union[User, Member]] = None  # documented as optional, for whatever reason
        self._permissions: int = 0

        # Store the locales
        self.user_locale = payload.get("locale")
        self.guild_locale = payload.get("guild_locale")

        # TODO: there's a potential data loss here
        if self.guild_id:
            guild = self.guild or Object(id=self.guild_id)
            try:
                member = payload['member']  # type: ignore
            except KeyError:
                pass
            else:
                self.user = Member(state=self._state, guild=guild, data=member)  # type: ignore
                self._permissions = int(member.get('permissions', 0))
        else:
            try:
                self.user = User(state=self._state, data=payload['user'])
            except KeyError:
                pass

    @property
    def locale(self) -> str:
        return self.user_locale or self.guild_locale or "en-US"

    @utils.cached_slot_property('_cs_resolved')
    def resolved(self) -> InteractionResolved:
        return InteractionResolved(interaction=self, data=(self.data or {}).copy().get("resolved", dict()), state=self._state)

    @utils.cached_slot_property('_cs_command_name')
    def command_name(self) -> Optional[str]:
        """Optional[:class:`str`]: The name of the invoked command. Only valid for slash command interactions."""

        if self.data is None:
            return
        data = self.data.copy()
        command_name = data.get('name')
        try:
            while command_name:
                if "options" not in data:
                    break
                if data['options'][0]['type'] in [1, 2]:
                    data = data['options'][0]
                    command_name += f" {data['name']}"
                else:
                    break
        finally:
            return command_name

    @property
    def guild(self) -> Optional[Guild]:
        """Optional[:class:`Guild`]: The guild the interaction was sent from."""
        return self._state and self._state._get_guild(self.guild_id)

    @utils.cached_slot_property('_cs_channel')
    def channel(self) -> Optional[InteractionChannel]:
        """Optional[Union[:class:`abc.GuildChannel`, :class:`PartialMessageable`, :class:`Thread`]]: The channel the interaction was sent from.

        Note that due to a Discord limitation, DM channels are not resolved since there is
        no data to complete them. These are :class:`PartialMessageable` instead.
        """
        guild = self.guild
        channel = guild and guild._resolve_channel(self.channel_id)
        if channel is None:
            if self.channel_id is not None:
                type = ChannelType.text if self.guild_id is not None else ChannelType.private
                return PartialMessageable(state=self._state, id=self.channel_id, type=type)
            return None
        return channel

    @property
    def permissions(self) -> Permissions:
        """:class:`Permissions`: The resolved permissions of the member in the channel, including overwrites.

        In a non-guild context where this doesn't apply, an empty permissions object is returned.
        """
        return Permissions(self._permissions)

    @utils.cached_slot_property('_cs_response')
    def response(self) -> Union[InteractionResponse, HTTPInteractionResponse]:
        """:class:`InteractionResponse`: Returns an object responsible for handling responding to the interaction.

        A response can only be done once. If secondary messages need to be sent, consider using :attr:`followup`
        instead.
        """
        return InteractionResponse(self)

    @utils.cached_slot_property('_cs_followup')
    def followup(self) -> Webhook:
        """:class:`Webhook`: Returns the follow up webhook for follow up interactions."""
        payload = {
            'id': self.application_id,
            'type': 3,
            'token': self.token,
        }
        v = Webhook.from_state(data=payload, state=self._state)
        v.supports_ephemeral = True
        return v

    async def original_message(self) -> InteractionMessage:
        """|coro|

        Fetches the original interaction response message associated with the interaction.

        If the interaction response was :meth:`InteractionResponse.send_message` then this would
        return the message that was sent using that response. Otherwise, this would return
        the message that triggered the interaction.

        Repeated calls to this will return a cached value.

        Raises
        -------
        HTTPException
            Fetching the original response message failed.
        ClientException
            The channel for the message could not be resolved.

        Returns
        --------
        InteractionMessage
            The original interaction response message.
        """

        if self._original_message is not None:
            return self._original_message

        # TODO: fix later to not raise?
        channel = self.channel
        if channel is None:
            raise ClientException('Channel for message could not be resolved')

        adapter = async_context.get()
        data = await adapter.get_original_interaction_response(
            application_id=self.application_id,
            token=self.token,
            session=self._session,
        )
        state = _InteractionMessageState(self, self._state)
        message = InteractionMessage(state=state, channel=channel, data=data)  # type: ignore
        self._original_message = message
        return message

    async def edit_original_message(
        self,
        *,
        content: Optional[str] = MISSING,
        embeds: List[Embed] = MISSING,
        embed: Optional[Embed] = MISSING,
        file: File = MISSING,
        files: List[File] = MISSING,
        components: Optional[MessageComponents] = MISSING,
        allowed_mentions: Optional[AllowedMentions] = None,
    ) -> InteractionMessage:
        """|coro|

        Edits the original interaction response message.

        This is a lower level interface to :meth:`InteractionMessage.edit` in case
        you do not want to fetch the message and save an HTTP request.

        This method is also the only way to edit the original message if
        the message sent was ephemeral.

        Parameters
        ------------
        content: Optional[:class:`str`]
            The content to edit the message with or ``None`` to clear it.
        embeds: List[:class:`Embed`]
            A list of embeds to edit the message with.
        embed: Optional[:class:`Embed`]
            The embed to edit the message with. ``None`` suppresses the embeds.
            This should not be mixed with the ``embeds`` parameter.
        file: :class:`File`
            The file to upload. This cannot be mixed with ``files`` parameter.
        files: List[:class:`File`]
            A list of files to send with the content. This cannot be mixed with the
            ``file`` parameter.
        allowed_mentions: :class:`AllowedMentions`
            Controls the mentions being processed in this message.
            See :meth:`.abc.Messageable.send` for more information.
        components: Optional[:class:`~discord.ui.MessageComponents`]
            The set of message components to update the message with. If ``None`` is passed then
            the components are removed.

        Raises
        -------
        HTTPException
            Editing the message failed.
        Forbidden
            Edited a message that is not yours.
        TypeError
            You specified both ``embed`` and ``embeds`` or ``file`` and ``files``
        ValueError
            The length of ``embeds`` was invalid.

        Returns
        --------
        :class:`InteractionMessage`
            The newly edited message.
        """

        previous_mentions: Optional[AllowedMentions] = self._state.allowed_mentions
        params = handle_message_parameters(
            content=content,
            file=file,
            files=files,
            embed=embed,
            embeds=embeds,
            components=components,
            allowed_mentions=allowed_mentions,
            previous_allowed_mentions=previous_mentions,
        )
        adapter = async_context.get()
        data = await adapter.edit_original_interaction_response(
            self.application_id,
            self.token,
            session=self._session,
            payload=params.payload,
            multipart=params.multipart,
            files=params.files,
        )

        # The message channel types should always match
        message = InteractionMessage(state=self._state, channel=self.channel, data=data)  # type: ignore
        return message

    async def delete_original_message(self) -> None:
        """|coro|

        Deletes the original interaction response message.

        This is a lower level interface to :meth:`InteractionMessage.delete` in case
        you do not want to fetch the message and save an HTTP request.

        Raises
        -------
        HTTPException
            Deleting the message failed.
        Forbidden
            Deleted a message that is not yours.
        """
        adapter = async_context.get()
        await adapter.delete_original_interaction_response(
            self.application_id,
            self.token,
            session=self._session,
        )


class InteractionResponse:
    """Represents a Discord interaction response.

    This type can be accessed through :attr:`Interaction.response`.
    """

    __slots__: Tuple[str, ...] = (
        '_responded',
        '_parent',
    )

    def __init__(self, parent: Interaction):
        self._parent: Interaction = parent
        self._responded: bool = False

    def is_done(self) -> bool:
        """:class:`bool`: Indicates whether an interaction response has been done before.

        An interaction can only be responded to once.
        """
        return self._responded

    async def defer(self, *, ephemeral: bool = False) -> None:
        """|coro|

        Defers the interaction response.

        This is typically used when the interaction is acknowledged
        and a secondary action will be done later.

        Parameters
        -----------
        ephemeral: :class:`bool`
            Indicates whether the deferred message will eventually be ephemeral.
            This only applies for interactions of type :attr:`InteractionType.application_command`.

        Raises
        -------
        HTTPException
            Deferring the interaction failed.
        InteractionResponded
            This interaction has already been responded to before.
        """
        if self._responded:
            raise InteractionResponded(self._parent)

        defer_type: int = 0
        data: Optional[Dict[str, Any]] = None
        parent = self._parent
        defer_type = InteractionResponseType.deferred_channel_message.value
        if ephemeral:
            data = {'flags': 64}

        if data:
            data = {
                "type": defer_type,
                "data": data,
            }
        else:
            data = {
                "type": defer_type,
            }

        if defer_type:
            adapter = async_context.get()
            await adapter.create_interaction_response(
                parent.id, parent.token, session=parent._session, payload=data
            )
            self._responded = True

    async def defer_update(self) -> None:
        """|coro|

        Defers the interaction response.

        This is typically used when the interaction is acknowledged
        and a secondary action will be done later.

        This is only usable if the interaction spawned from a component interaction.

        Raises
        -------
        HTTPException
            Deferring the interaction failed.
        InteractionResponded
            This interaction has already been responded to before.
        """
        if self._responded:
            raise InteractionResponded(self._parent)

        defer_type: int = 0
        data: Optional[Dict[str, Any]] = None
        parent = self._parent
        defer_type = InteractionResponseType.deferred_message_update.value

        if defer_type:
            payload = {"type": defer_type}
            adapter = async_context.get()
            await adapter.create_interaction_response(
                parent.id, parent.token, session=parent._session, payload=payload
            )
            self._responded = True

    async def pong(self) -> None:
        """|coro|

        Pongs the ping interaction.

        This should rarely be used.

        Raises
        -------
        HTTPException
            Ponging the interaction failed.
        InteractionResponded
            This interaction has already been responded to before.
        """
        if self._responded:
            raise InteractionResponded(self._parent)

        data = {
            "type": InteractionResponseType.pong.value,
        }

        parent = self._parent
        if parent.type is InteractionType.ping:
            adapter = async_context.get()
            await adapter.create_interaction_response(
                parent.id, parent.token, session=parent._session, payload=data,
            )
            self._responded = True

    async def send_autocomplete(
        self,
        options: List[ApplicationCommandOptionChoice] = None,
    ) -> None:
        """|coro|

        Responds to this interaction by sending a message.

        .. versionadded:: 0.0.4

        Parameters
        -----------
        options: typing.List[:class:`ApplicationCommandOptionChoice`]
            The options that you want to send back to Discord.

        Raises
        -------
        HTTPException
            Sending the message failed.
        ValueError
            The length of ``options`` was invalid.
        InteractionResponded
            This interaction has already been responded to before.
        """
        if self._responded:
            raise InteractionResponded(self._parent)

        parent = self._parent
        payload = {
            "type": InteractionResponseType.autocomplete.value,
            "data": {
                "choices": [i.to_json() for i in options or list()],
            },
        }

        adapter = async_context.get()
        await adapter.create_interaction_response(
            parent.id,
            parent.token,
            session=parent._session,
            payload=payload,
        )

        self._responded = True

    async def send_modal(
        self,
        modal: Modal,
    ) -> None:
        """|coro|

        Reponds to the interaction by sending a modal back to the user.

        .. versionadded:: 0.0.5

        Parameters
        -----------
        modal: Modal
            The modal that you want to send to the user.

        Raises
        -------
        HTTPException
            Sending the message failed.
        InteractionResponded
            This interaction has already been responded to before.
        """
        if self._responded:
            raise InteractionResponded(self._parent)

        parent = self._parent
        payload = {
            "type": InteractionResponseType.modal.value,
            "data": modal.to_dict(),
        }

        adapter = async_context.get()
        await adapter.create_interaction_response(
            parent.id,
            parent.token,
            session=parent._session,
            payload=payload,
        )

        self._responded = True

    async def send_message(
        self,
        content: Optional[Any] = None,
        *,
        embed: Embed = MISSING,
        embeds: List[Embed] = MISSING,
        file: File = MISSING,
        files: List[File] = MISSING,
        components: MessageComponents = MISSING,
        tts: bool = False,
        ephemeral: bool = False,
        allowed_mentions: AllowedMentions = None,
    ) -> None:
        """|coro|

        Responds to this interaction by sending a message.

        Parameters
        -----------
        content: Optional[:class:`str`]
            The content of the message to send.
        embeds: List[:class:`Embed`]
            A list of embeds to send with the content. Maximum of 10. This cannot
            be mixed with the ``embed`` parameter.
        embed: :class:`Embed`
            The rich embed for the content to send. This cannot be mixed with
            ``embeds`` parameter.
        file: :class:`File`
            The file to upload. This cannot be mixed with ``files`` parameter.
        files: List[:class:`File`]
            A list of files to send with the content. This cannot be mixed with the
            ``file`` parameter.
        tts: :class:`bool`
            Indicates if the message should be sent using text-to-speech.
        components: :class:`discord.ui.MessageComponents`
            The components to send with the messasge.
        ephemeral: :class:`bool`
            Indicates if the message should only be visible to the user who started the interaction.
        allowed_mentions: :class:`AllowedMentions`
            The allowed mentions that should be sent with the message.

        Raises
        -------
        HTTPException
            Sending the message failed.
        TypeError
            You specified both ``embed`` and ``embeds``.
        ValueError
            The length of ``embeds`` was invalid.
        InteractionResponded
            This interaction has already been responded to before.
        """
        if self._responded:
            raise InteractionResponded(self._parent)

        parent = self._parent
        params = handle_message_parameters(
            content=content,
            tts=tts,
            file=file,
            files=files,
            embed=embed,
            embeds=embeds,
            ephemeral=ephemeral,
            components=components,
            allowed_mentions=allowed_mentions,
            previous_allowed_mentions=parent._state.allowed_mentions,
            type=InteractionResponseType.channel_message.value,
        )

        adapter = async_context.get()
        await adapter.create_interaction_response(
            parent.id,
            parent.token,
            session=parent._session,
            payload=params.payload,
            multipart=params.multipart,
            files=params.files,
        )

        self._responded = True

    async def edit_message(
        self,
        *,
        content: Optional[Any] = MISSING,
        embed: Optional[Embed] = MISSING,
        embeds: List[Embed] = MISSING,
        attachments: List[Attachment] = MISSING,
        components: Optional[MessageComponents] = MISSING,
        allowed_mentions: Optional[AllowedMentions] = MISSING,
    ) -> None:
        """|coro|

        Responds to this interaction by editing the original message of
        a component interaction.

        Parameters
        -----------
        content: Optional[:class:`str`]
            The new content to replace the message with. ``None`` removes the content.
        embeds: List[:class:`Embed`]
            A list of embeds to edit the message with.
        embed: Optional[:class:`Embed`]
            The embed to edit the message with. ``None`` suppresses the embeds.
            This should not be mixed with the ``embeds`` parameter.
        attachments: List[:class:`Attachment`]
            A list of attachments to keep in the message. If ``[]`` is passed
            then all attachments are removed.
        components: Optional[:class:`~discord.ui.MessageComponents`]
            The updated components to update this message with. If ``None`` is passed then
            the components are removed.

        Raises
        -------
        HTTPException
            Editing the message failed.
        TypeError
            You specified both ``embed`` and ``embeds``.
        InteractionResponded
            This interaction has already been responded to before.
        """
        if self._responded:
            raise InteractionResponded(self._parent)

        parent = self._parent
        msg = parent.message
        state = parent._state
        message_id = msg.id if msg else None
        if parent.type is not InteractionType.component:
            return

        payload = {}
        if content is not MISSING:
            if content is None:
                payload['content'] = None
            else:
                payload['content'] = str(content)

        if embed is not MISSING and embeds is not MISSING:
            raise TypeError('cannot mix both embed and embeds keyword arguments')

        if embed is not MISSING:
            if embed is None:
                embeds = []
            else:
                embeds = [embed]

        if embeds is not MISSING:
            payload['embeds'] = [e.to_dict() for e in embeds]

        if attachments is not MISSING:
            payload['attachments'] = [a.to_dict() for a in attachments]

        if components is not MISSING:
            if components is None:
                payload['components'] = []
            else:
                payload['components'] = components.to_dict()

        if allowed_mentions is not MISSING:
            parent_state = self._parent._state
            if allowed_mentions:
                if parent_state.allowed_mentions is not None:
                    payload['allowed_mentions'] = parent_state.allowed_mentions.merge(allowed_mentions).to_dict()
                else:
                    payload['allowed_mentions'] = allowed_mentions.to_dict()
            elif parent_state.allowed_mentions is not None:
                payload['allowed_mentions'] = parent_state.allowed_mentions.to_dict()

        payload = {
            "type": InteractionResponseType.message_update.value,
            "data": payload,
        }

        adapter = async_context.get()
        await adapter.create_interaction_response(
            parent.id,
            parent.token,
            session=parent._session,
            payload=payload,
        )

        self._responded = True


class _MultipartWriter(object):

    def __init__(self):
        self.buffer = bytearray()

    async def write(self, data):
        self.buffer.extend(data)


class HTTPInteractionResponse(InteractionResponse):
    """Represents a Discord interaction response.

    This type can be accessed through :attr:`Interaction.response`.
    """

    __slots__: Tuple[str, ...] = (
        '_responded',
        '_parent',
        '_aiohttp_request',
        '_aiohttp_response',
    )

    def __init__(self, aiohttp_response: web.Request, parent: Interaction):
        super().__init__(parent)
        self._aiohttp_request: web.Request = aiohttp_response
        self._aiohttp_response: web.StreamResponse = web.StreamResponse()

    async def defer(self, *, ephemeral: bool = False) -> None:
        """|coro|

        Defers the interaction response.

        This is typically used when the interaction is acknowledged
        and a secondary action will be done later.

        Parameters
        -----------
        ephemeral: :class:`bool`
            Indicates whether the deferred message will eventually be ephemeral.
            This only applies for interactions of type :attr:`InteractionType.application_command`.

        Raises
        -------
        HTTPException
            Deferring the interaction failed.
        InteractionResponded
            This interaction has already been responded to before.
        """
        if self._responded:
            raise InteractionResponded(self._parent)

        defer_type: int = 0
        data: Optional[Dict[str, Any]] = None
        defer_type = InteractionResponseType.deferred_channel_message.value
        if ephemeral:
            data = {'flags': 64}

        if defer_type:
            payload = {"type": defer_type}
            if data:
                payload["data"] = data
            self._aiohttp_response.headers["Content-Type"] = "application/json"
            await self._aiohttp_response.prepare(self._aiohttp_request)
            await self._aiohttp_response.write(json.dumps(payload).encode())
            await self._aiohttp_response.write_eof()
            self._responded = True

    async def defer_update(self) -> None:
        """|coro|

        Defers the interaction response.

        This is typically used when the interaction is acknowledged
        and a secondary action will be done later.

        This is only usable if the interaction spawned from a component interaction.

        Raises
        -------
        HTTPException
            Deferring the interaction failed.
        InteractionResponded
            This interaction has already been responded to before.
        """
        if self._responded:
            raise InteractionResponded(self._parent)

        defer_type: int = 0
        defer_type = InteractionResponseType.deferred_message_update.value

        if defer_type:
            payload = {"type": defer_type}
            self._aiohttp_response.headers["Content-Type"] = "application/json"
            await self._aiohttp_response.prepare(self._aiohttp_request)
            await self._aiohttp_response.write(json.dumps(payload).encode())
            await self._aiohttp_response.write_eof()
            self._responded = True

    async def pong(self) -> None:
        """|coro|

        Pongs the ping interaction.

        This should rarely be used.

        Raises
        -------
        HTTPException
            Ponging the interaction failed.
        InteractionResponded
            This interaction has already been responded to before.
        """
        if self._responded:
            raise InteractionResponded(self._parent)

        parent = self._parent
        if parent.type is InteractionType.ping:
            payload = {"type": InteractionResponseType.pong.value}
            self._aiohttp_response.headers["Content-Type"] = "application/json"
            await self._aiohttp_response.prepare(self._aiohttp_request)
            await self._aiohttp_response.write(json.dumps(payload).encode())
            await self._aiohttp_response.write_eof()
            self._responded = True

    async def send_message(
        self,
        content: Optional[Any] = None,
        *,
        embed: Embed = MISSING,
        embeds: List[Embed] = MISSING,
        file: File = MISSING,
        files: List[File] = MISSING,
        components: MessageComponents = MISSING,
        tts: bool = False,
        ephemeral: bool = False,
        allowed_mentions: AllowedMentions = None,
    ) -> None:
        """|coro|

        Responds to this interaction by sending a message.

        Parameters
        -----------
        content: Optional[:class:`str`]
            The content of the message to send.
        embeds: List[:class:`Embed`]
            A list of embeds to send with the content. Maximum of 10. This cannot
            be mixed with the ``embed`` parameter.
        embed: :class:`Embed`
            The rich embed for the content to send. This cannot be mixed with
            ``embeds`` parameter.
        file: :class:`File`
            The file to upload. This cannot be mixed with ``files`` parameter.
        files: List[:class:`File`]
            A list of files to send with the content. This cannot be mixed with the
            ``file`` parameter.
        tts: :class:`bool`
            Indicates if the message should be sent using text-to-speech.
        components: :class:`discord.ui.MessageComponents`
            The components to send with the messasge.
        ephemeral: :class:`bool`
            Indicates if the message should only be visible to the user who started the interaction.
        allowed_mentions: :class:`AllowedMentions`
            The allowed mentions that should be sent with the message.

        Raises
        -------
        HTTPException
            Sending the message failed.
        TypeError
            You specified both ``embed`` and ``embeds``.
        ValueError
            The length of ``embeds`` was invalid.
        InteractionResponded
            This interaction has already been responded to before.
        """
        if self._responded:
            raise InteractionResponded(self._parent)

        params = handle_message_parameters(
            content=content,
            tts=tts,
            file=file,
            files=files,
            embed=embed,
            embeds=embeds,
            ephemeral=ephemeral,
            components=components,
            allowed_mentions=allowed_mentions,
            previous_allowed_mentions=self._parent._state.allowed_mentions,
            type=InteractionResponseType.channel_message.value,
        )

        payload = params.payload
        multipart = params.multipart
        files = params.files

        headers = {}
        to_send = None
        if payload is not None:
            headers['Content-Type'] = 'application/json'
            to_send = utils._to_json(payload).encode()

        for file in files or list():
            file.reset(seek=0)

        if multipart:
            form_data = aiohttp.FormData()
            for p in multipart:
                form_data.add_field(**p)
            writer = _MultipartWriter()
            payload = form_data()
            await payload.write(writer)
            to_send = writer.buffer
            headers = payload.headers

        assert to_send

        self._aiohttp_response.headers.update(headers)
        await self._aiohttp_response.prepare(self._aiohttp_request)
        await self._aiohttp_response.write(to_send)
        await self._aiohttp_response.write_eof()

        self._responded = True

    async def send_autocomplete(
        self,
        options: List[ApplicationCommandOptionChoice] = None,
    ) -> None:
        """|coro|

        Responds to this interaction by sending a message.

        Parameters
        -----------
        options: typing.List[:class:`ApplicationCommandOptionChoice`]
            The options that you want to send back to Discord.

        Raises
        -------
        HTTPException
            Sending the message failed.
        ValueError
            The length of ``options`` was invalid.
        InteractionResponded
            This interaction has already been responded to before.
        """
        if self._responded:
            raise InteractionResponded(self._parent)

        parent = self._parent
        payload = {
            "type": InteractionResponseType.autocomplete.value,
            "data": {
                "choices": [i.to_json() for i in options or list()],
            },
        }

        self._aiohttp_response.headers["Content-Type"] = "application/json"
        await self._aiohttp_response.prepare(self._aiohttp_request)
        await self._aiohttp_response.write(json.dumps(payload).encode())
        await self._aiohttp_response.write_eof()

        self._responded = True

    async def send_modal(
        self,
        modal: Modal,
    ) -> None:
        """|coro|

        Reponds to the interaction by sending a modal back to the user.

        .. versionadded:: 0.0.5

        Parameters
        -----------
        modal: Modal
            The modal that you want to send to the user.

        Raises
        -------
        HTTPException
            Sending the message failed.
        InteractionResponded
            This interaction has already been responded to before.
        """
        if self._responded:
            raise InteractionResponded(self._parent)

        parent = self._parent
        payload = {
            "type": InteractionResponseType.modal.value,
            "data": modal.to_dict(),
        }

        self._aiohttp_response.headers["Content-Type"] = "application/json"
        await self._aiohttp_response.prepare(self._aiohttp_request)
        await self._aiohttp_response.write(json.dumps(payload).encode())
        await self._aiohttp_response.write_eof()

        self._responded = True

    async def edit_message(
        self,
        *,
        content: Optional[Any] = MISSING,
        embed: Optional[Embed] = MISSING,
        embeds: List[Embed] = MISSING,
        attachments: List[Attachment] = MISSING,
        components: Optional[MessageComponents] = MISSING,
        allowed_mentions: Optional[AllowedMentions] = MISSING,
    ) -> None:
        """|coro|

        Responds to this interaction by editing the original message of
        a component interaction.

        Parameters
        -----------
        content: Optional[:class:`str`]
            The new content to replace the message with. ``None`` removes the content.
        embeds: List[:class:`Embed`]
            A list of embeds to edit the message with.
        embed: Optional[:class:`Embed`]
            The embed to edit the message with. ``None`` suppresses the embeds.
            This should not be mixed with the ``embeds`` parameter.
        attachments: List[:class:`Attachment`]
            A list of attachments to keep in the message. If ``[]`` is passed
            then all attachments are removed.
        components: Optional[:class:`~discord.ui.MessageComponents`]
            The updated components to update this message with. If ``None`` is passed then
            the components are removed.

        Raises
        -------
        HTTPException
            Editing the message failed.
        TypeError
            You specified both ``embed`` and ``embeds``.
        InteractionResponded
            This interaction has already been responded to before.
        """
        if self._responded:
            raise InteractionResponded(self._parent)

        parent = self._parent
        msg = parent.message
        state = parent._state
        message_id = msg.id if msg else None
        if parent.type is not InteractionType.component:
            return

        payload = {}
        if content is not MISSING:
            if content is None:
                payload['content'] = None
            else:
                payload['content'] = str(content)

        if embed is not MISSING and embeds is not MISSING:
            raise TypeError('cannot mix both embed and embeds keyword arguments')

        if embed is not MISSING:
            if embed is None:
                embeds = []
            else:
                embeds = [embed]

        if embeds is not MISSING:
            payload['embeds'] = [e.to_dict() for e in embeds]

        if attachments is not MISSING:
            payload['attachments'] = [a.to_dict() for a in attachments]

        if components is not MISSING:
            if components is None:
                payload['components'] = []
            else:
                payload['components'] = components.to_dict()

        if allowed_mentions is not MISSING:
            parent_state = self._parent._state
            if allowed_mentions:
                if parent_state.allowed_mentions is not None:
                    payload['allowed_mentions'] = parent_state.allowed_mentions.merge(allowed_mentions).to_dict()
                else:
                    payload['allowed_mentions'] = allowed_mentions.to_dict()
            elif parent_state.allowed_mentions is not None:
                payload['allowed_mentions'] = parent_state.allowed_mentions.to_dict()

        data = {
            "type": InteractionResponseType.message_update.value,
            "data": payload,
        }
        self._aiohttp_response.headers["Content-Type"] = "application/json"
        await self._aiohttp_response.prepare(self._aiohttp_request)
        await self._aiohttp_response.write(json.dumps(data).encode())
        await self._aiohttp_response.write_eof()

        self._responded = True


class _InteractionMessageState:
    __slots__ = ('_parent', '_interaction')

    def __init__(self, interaction: Interaction, parent: ConnectionState):
        self._interaction: Interaction = interaction
        self._parent: ConnectionState = parent

    def _get_guild(self, guild_id):
        return self._parent._get_guild(guild_id)

    def store_user(self, data):
        return self._parent.store_user(data)

    def create_user(self, data):
        return self._parent.create_user(data)

    @property
    def http(self):
        return self._parent.http

    def __getattr__(self, attr):
        return getattr(self._parent, attr)


class InteractionMessage(Message):
    """Represents the original interaction response message.

    This allows you to edit or delete the message associated with
    the interaction response. To retrieve this object see :meth:`Interaction.original_message`.

    This inherits from :class:`discord.Message` with changes to
    :meth:`edit` and :meth:`delete` to work.
    """

    __slots__ = ()
    _state: _InteractionMessageState

    async def edit(
        self,
        content: Optional[str] = MISSING,
        embeds: List[Embed] = MISSING,
        embed: Optional[Embed] = MISSING,
        file: File = MISSING,
        files: List[File] = MISSING,
        components: Optional[MessageComponents] = MISSING,
        allowed_mentions: Optional[AllowedMentions] = None,
    ) -> InteractionMessage:
        """|coro|

        Edits the message.

        Parameters
        ------------
        content: Optional[:class:`str`]
            The content to edit the message with or ``None`` to clear it.
        embeds: List[:class:`Embed`]
            A list of embeds to edit the message with.
        embed: Optional[:class:`Embed`]
            The embed to edit the message with. ``None`` suppresses the embeds.
            This should not be mixed with the ``embeds`` parameter.
        file: :class:`File`
            The file to upload. This cannot be mixed with ``files`` parameter.
        files: List[:class:`File`]
            A list of files to send with the content. This cannot be mixed with the
            ``file`` parameter.
        allowed_mentions: :class:`AllowedMentions`
            Controls the mentions being processed in this message.
            See :meth:`.abc.Messageable.send` for more information.
        components: Optional[:class:`~discord.ui.MessageComponents`]
            The updated components to update this message with. If ``None`` is passed then
            the components are removed.

        Raises
        -------
        HTTPException
            Editing the message failed.
        Forbidden
            Edited a message that is not yours.
        TypeError
            You specified both ``embed`` and ``embeds`` or ``file`` and ``files``
        ValueError
            The length of ``embeds`` was invalid.

        Returns
        ---------
        :class:`InteractionMessage`
            The newly edited message.
        """
        return await self._state._interaction.edit_original_message(
            content=content,
            embeds=embeds,
            embed=embed,
            file=file,
            files=files,
            components=components,
            allowed_mentions=allowed_mentions,
        )

    async def delete(self, *, delay: Optional[float] = None) -> None:
        """|coro|

        Deletes the message.

        Parameters
        -----------
        delay: Optional[:class:`float`]
            If provided, the number of seconds to wait before deleting the message.
            The waiting is done in the background and deletion failures are ignored.

        Raises
        ------
        Forbidden
            You do not have proper permissions to delete the message.
        NotFound
            The message was deleted already.
        HTTPException
            Deleting the message failed.
        """

        if delay is not None:

            async def inner_call(delay: float = delay):
                await asyncio.sleep(delay)
                try:
                    await self._state._interaction.delete_original_message()
                except HTTPException:
                    pass

            asyncio.create_task(inner_call())
        else:
            await self._state._interaction.delete_original_message()



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .types.components import (
        Component as ComponentPayload,
    )


def _component_factory(data: ComponentPayload):
    from .ui import (
        BaseComponent,  # Fallback
        ActionRow,  # 1
        Button,  # 2
        SelectMenu,  # 3
        # Modal is 4 but that'll never appear on a message, and shouldn't be sent to a bot
        UserSelectMenu,  # 5
        RoleSelectMenu,  # 6
        MentionableSelectMenu,  # 7
        ChannelSelectMenu,  # 8
    )
    component_type = data['type']
    if component_type == 1:
        return ActionRow.from_dict(data)
    elif component_type == 2:
        return Button.from_dict(data)
    elif component_type == 3:
        return SelectMenu.from_dict(data)
    elif component_type == 5:
        return UserSelectMenu.from_dict(data)
    elif component_type == 6:
        return RoleSelectMenu.from_dict(data)
    elif component_type == 7:
        return MentionableSelectMenu.from_dict(data)
    elif component_type == 8:
        return ChannelSelectMenu.from_dict(data)
    else:
        v = BaseComponent()
        v.data = data
        v.custom_id = data.get("custom_id")
        return v



import argparse
import typing
import pathlib
import textwrap

from .runner import run_bot, run_website, run_sharder, run_shell, run_modify_commands, run_interactions


def get_path_relative_to_file(
        path: pathlib.Path
        ) -> pathlib.Path:
    """
    Get the full path for a file relative to a given location.
    """

    here = pathlib.Path(__file__).parent.resolve()
    return here.joinpath(path)


def create_file(*path: str, content: typing.Optional[typing.Union[str, pathlib.Path]] = None):
    joined_folder_path = pathlib.Path("./").joinpath(*path[:-1])
    joined_file_path = pathlib.Path("./").joinpath(*path)
    if content:
        if isinstance(content, pathlib.Path):
            filename = content
            with open(filename) as a:
                content = a.read()
        joined_folder_path.mkdir(parents=True, exist_ok=True)
        try:
            with open(joined_file_path, "x") as a:
                a.write(content)
        except FileExistsError:
            print(f"File {joined_file_path} was not created as one already exists.")
    else:
        joined_file_path.mkdir(parents=True, exist_ok=True)


def get_default_program_arguments() -> argparse.ArgumentParser:
    """
    Set up the program arguments for the module. These include the following (all are proceeded by "python -m voxelbotutils"):
    "run bot config.toml --min 0 --max 10 --shardcount 10"
    "run bot config/config.toml"
    "run website config.toml"
    "run website config/config.toml"
    "create-config bot"
    "create-config website"

    Returns
    --------
    :class:`argparse.ArgumentParser`
        The arguments that were parsed
    """

    LOGLEVEL_CHOICES = ["debug", "info", "warning", "error", "critical"]

    # Set up our parsers and subparsers
    parser = argparse.ArgumentParser(prog="vbu")
    runner_subparser = parser.add_subparsers(dest="subcommand")
    runner_subparser.required = True
    bot_subparser = runner_subparser.add_parser("run-bot")
    website_subparser = runner_subparser.add_parser("run-website")
    interactions_subparser = runner_subparser.add_parser("run-interactions")
    sharder_subparser = runner_subparser.add_parser("run-sharder")
    shell_subparser = runner_subparser.add_parser("run-shell")
    application_subparser = runner_subparser.add_parser("commands")
    create_config_subparser = runner_subparser.add_parser("create-config")
    check_config_subparser = runner_subparser.add_parser("check-config")
    runner_subparser.add_parser("version")

    # Set up the bot arguments
    bot_subparser.add_argument("bot_directory", nargs="?", default=".", help="The directory containing a config and a cogs folder for the bot to run.")
    bot_subparser.add_argument("config_file", nargs="?", default="config/config.toml", help="The configuration for the bot.")
    bot_subparser.add_argument("--min", nargs="?", type=int, default=None, help="The minimum shard ID that this instance will run with (inclusive).")
    bot_subparser.add_argument("--max", nargs="?", type=int, default=None, help="The maximum shard ID that this instance will run with (inclusive).")
    bot_subparser.add_argument("--shardcount", nargs="?", type=int, default=1, help="The amount of shards that the bot should be using.")
    bot_subparser.add_argument("--loglevel", nargs="?", default="INFO", help="Global logging level - probably most useful is INFO and DEBUG.", choices=LOGLEVEL_CHOICES)

    # Set up the website arguments
    website_subparser.add_argument("website_directory", nargs="?", default=".", help="The directory containing a static and templates folder for the website to run.")
    website_subparser.add_argument("config_file", nargs="?", default="config/website.toml", help="The configuration for the website.")
    website_subparser.add_argument("--host", nargs="?", default="0.0.0.0", help="The host IP to run the website on.")
    website_subparser.add_argument("--port", nargs="?", type=int, default="8080", help="The port to run the website with.")
    website_subparser.add_argument("--debug", action="store_true", default=False, help="Whether or not to run the website in debug mode.")
    website_subparser.add_argument("--loglevel", nargs="?", default="INFO", help="Global logging level - probably most useful is INFO and DEBUG.", choices=LOGLEVEL_CHOICES)

    # Set up the interactions server arguments
    interactions_subparser.add_argument("bot_directory", nargs="?", default=".", help="The directory containing a config and a cogs folder for the bot to run.")
    interactions_subparser.add_argument("config_file", nargs="?", default="config/config.toml", help="The configuration for the bot.")
    interactions_subparser.add_argument("--host", nargs="?", default="0.0.0.0", help="The host IP to run the website on.")
    interactions_subparser.add_argument("--port", nargs="?", type=int, default="8080", help="The port to run the website with.")
    interactions_subparser.add_argument("--path", nargs="?", type=str, default="/interactions", help="The path to run the interactions endpoint on.")
    interactions_subparser.add_argument("--connect", action="store_true", default=False, help="Whether you want your bot to connect to the Discord gateway.")
    interactions_subparser.add_argument("--debug", action="store_true", default=False, help="Whether or not to run the website in debug mode.")
    interactions_subparser.add_argument("--loglevel", nargs="?", default="INFO", help="Global logging level - probably most useful is INFO and DEBUG.", choices=LOGLEVEL_CHOICES)

    # Set up the sharder arguments
    sharder_subparser.add_argument("config_file", nargs="?", default="config/config.toml", help="The configuration for the bot.")
    sharder_subparser.add_argument("--host", nargs="?", default="127.0.0.1", help="The host address to listen on.")
    sharder_subparser.add_argument("--port", nargs="?", default=8888, type=int, help="The host port to listen on.")
    sharder_subparser.add_argument("--concurrency", nargs="?", default=1, type=int, help="The max concurrency of the connecting bot.")
    sharder_subparser.add_argument("--loglevel", nargs="?", default="INFO", help="Global logging level - probably most useful is INFO and DEBUG.", choices=LOGLEVEL_CHOICES)

    # Set up the shell arguments
    shell_subparser.add_argument("bot_directory", nargs="?", default=".", help="The directory containing a config and a cogs folder for the bot to run.")
    shell_subparser.add_argument("config_file", nargs="?", default="config/config.toml", help="The configuration for the bot.")

    # Set up the shell arguments
    application_subparser.add_argument("action", help="The action yuou want to take on your application commands.", choices=["add", "remove"])
    application_subparser.add_argument("bot_directory", nargs="?", default=".", help="The directory containing a config and a cogs folder for the bot to run.")
    application_subparser.add_argument("config_file", nargs="?", default="config/config.toml", help="The configuration for the bot.")
    application_subparser.add_argument("--guild", nargs="?", default=None, type=int, help="The guild that you want to modify the application commands for.")

    # See what we want to make a config file for
    create_config_subparser.add_argument("config_type", nargs=1, help="The type of config file that we want to create.", choices=["bot", "website", "all"])
    check_config_subparser.add_argument("config_type", nargs=1, help="The type of config file that we want to create.", choices=["bot", "website"])
    check_config_subparser.add_argument("config_file", nargs="?", default="config/config.toml", help="The configuration file that you want to check.")

    # Wew that's a lot of things
    return parser


def check_config_value(
        base_config_key: typing.List[str], base_config_value: typing.Any,
        compare_config_value: typing.Any) -> None:
    """
    Recursively checks a config item to see if it's valid against a base config item
    """

    default_value = base_config_value if not isinstance(base_config_value, str) else textwrap.dedent(base_config_value).replace("\n", "\\n")

    # See if the item was omitted
    if isinstance(compare_config_value, type(None)):
        print(f"No value {base_config_key} was provided in your config file - should be type {type(base_config_value).__name__} (default {default_value!r}).")
        if isinstance(base_config_value, dict):
            for i, o in base_config_value.items():
                check_config_value(base_config_key + [i], o, None)
        return

    # See if the item was a str when it should be something else
    if not isinstance(base_config_value, type(compare_config_value)):
        print(f"Wrong value {base_config_key} type was provided in your config file - should be type {type(base_config_value).__name__} (default {default_value!r}).")
        if isinstance(base_config_value, dict):
            for i, o in base_config_value.items():
                check_config_value(base_config_key + [i], o, None)
        return

    # Correct value type - see if it was dict and recurse
    if isinstance(base_config_value, dict):
        for i, o in base_config_value.items():
            check_config_value(base_config_key + [i], o, compare_config_value.get(i))
    return


def get_next_version(current_version) -> str:
    """
    Get the next version that we can use in the requirements file.
    """

    release, major, minor = [int("".join(o for o in i if o.isdigit())) for i in current_version.split(".")]
    if release == 0:
        return f"0.{major + 1}.0"
    return f"{release + 1}.0.0"


def main():
    """
    The main method for running all of vbu.
    """

    # Wew let's see if we want to run a bot
    parser = get_default_program_arguments()
    args = parser.parse_args()
    from . import __version__
    next_version = get_next_version(__version__)

    # Let's see if we copyin bois
    if args.subcommand == "create-config":
        config_type = args.config_type[0]
        if config_type in ["website", "all"]:
            create_file("config", "website.toml", content=get_path_relative_to_file("config/web_config_example_file.toml"))
            create_file("config", "website.example.toml", content=get_path_relative_to_file("config/web_config_example_file.toml"))
            create_file("config", "database.pgsql", content=get_path_relative_to_file("config/database_base_file.pgsql"))
            create_file("_run_website.sh", content="vbu run-website .\n")
            create_file(".gitignore", content="__pycache__/\n.venv/\nconfig/config.toml\nconfig/website.toml\n")
            create_file("requirements.txt", content=f"voxelbotutils[web]>={__version__},<{next_version}\n")
            create_file("website", "frontend.py", content=get_path_relative_to_file("config/website_frontend_content.py"))
            create_file("website", "backend.py", content=get_path_relative_to_file("config/website_backend_content.py"))
            create_file("website", "static", ".gitkeep", content="\n")
            create_file("website", "templates", ".gitkeep", content="\n")
            create_file(".venv")
            print("Created website config file.")
        if config_type in ["bot", "all"]:
            create_file("config", "config.toml", content=get_path_relative_to_file("config/config_example_file.toml"))
            create_file("config", "config.example.toml", content=get_path_relative_to_file("config/config_example_file.toml"))
            create_file("config", "database.pgsql", content=get_path_relative_to_file("config/database_base_file.pgsql"))
            create_file("cogs", "ping_command.py", content=get_path_relative_to_file("config/cog_example_file.py"))
            create_file("_run_bot.sh", content="vbu run-bot .\n")
            create_file(".gitignore", content="__pycache__/\nconfig/config.toml\nconfig/website.toml\n")
            create_file("requirements.txt", content=f"voxelbotutils>={__version__},<{next_version}\n")
            create_file(".venv")
            print("Created bot config file.")
        exit(1)

    # See if we want to check the config file
    elif args.subcommand == "check-config":
        config_type = args.config_type[0]
        import toml
        if config_type == "website":
            with open(get_path_relative_to_file("config/web_config_example_file.toml")) as a:
                base_config_file_text = a.read()
        elif config_type == "bot":
            with open(get_path_relative_to_file("config/config_example_file.toml")) as a:
                base_config_file_text = a.read()
        else:
            exit(1)
        base_config_file = toml.loads(base_config_file_text)
        try:
            with open(args.config_file) as a:
                compare_config_file = toml.load(a)
        except Exception:
            print(f"Couldn't open config file {args.config_file}")
            exit(0)
        for key, value in base_config_file.items():
            check_config_value([key], value, compare_config_file.get(key))
        print("Completed config check.")
        exit(1)

    # If we just want the version
    elif args.subcommand == "version":
        print(f"VoxelBotUtils v{__version__}")
        exit(1)

    # Run things
    elif args.subcommand == "run-bot":
        run_bot(args)
    elif args.subcommand == "run-interactions":
        run_interactions(args)
    elif args.subcommand == "run-website":
        run_website(args)
    elif args.subcommand == "run-sharder":
        run_sharder(args)
    elif args.subcommand == "run-shell":
        run_shell(args)
    elif args.subcommand == "commands":
        run_modify_commands(args)


if __name__ == '__main__':
    main()
